# This file is part of the MapProxy project.
# Copyright (C) 2010 Omniscale <http://omniscale.de>
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import print_function


import re
import threading
import sys
import socket
import errno
import time
import base64
from contextlib import contextmanager
from urllib.parse import parse_qsl, urlsplit

from mapproxy.util.py import reraise
from http.server import HTTPServer as HTTPServer_, BaseHTTPRequestHandler


class RequestsMismatchError(AssertionError):
    def __init__(self, assertions):
        self.assertions = assertions

    def __str__(self):
        assertions = []
        for assertion in self.assertions:
            assertions.append(text_indent(str(assertion), '    ', ' -  '))
        return 'requests mismatch:\n' + '\n'.join(assertions)


class RequestError(str):
    pass


def text_indent(text, indent, first_indent=None):
    if first_indent is None:
        first_indent = indent

    text = first_indent + text
    return text.replace('\n', '\n' + indent)


class RequestMismatch(object):
    def __init__(self, msg, expected, actual):
        self.msg = msg
        self.expected = expected
        self.actual = actual

    def __str__(self):
        return ('requests mismatch (%s), expected:\n' % self.msg +
                text_indent(str(self.expected), '    ') +
                '\n  got:\n' + text_indent(str(self.actual), '    '))


class HTTPServer(HTTPServer_):
    allow_reuse_address = True

    def handle_error(self, request, client_address):
        _exc_class, exc, _tb = sys.exc_info()
        if isinstance(exc, socket.error):
            if exc.errno == errno.EPIPE:
                # suppres 'Broken pipe' errors raised in timeout tests
                return
        HTTPServer_.handle_error(self, request, client_address)


class ThreadedStopableHTTPServer(threading.Thread):
    def __init__(self, address, requests_responses, unordered=False, query_comparator=None):
        threading.Thread.__init__(self, **{'group': None})
        self.requests_responses = requests_responses
        self.daemon = True
        self.sucess = False
        self.shutdown = False
        self.httpd = HTTPServer(address, mock_http_handler(requests_responses,
                                                           unordered=unordered, query_comparator=query_comparator))
        self.httpd.timeout = 1.0
        self.assertions = self.httpd.assertions = []

    @property
    def http_port(self):
        return self.httpd.socket.getsockname()[1]

    def run(self):
        while self.requests_responses:
            if self.shutdown:
                break
            self.httpd.handle_request()
        if self.requests_responses:
            missing_req = [req for req, resp in self.requests_responses]
            self.assertions.append(
                RequestError('missing requests: ' + ','.join(map(str, missing_req)))
            )
        if not self.assertions:
            self.sucess = True
        # force socket close so next test can bind to same address
        self.httpd.socket.close()


class ThreadedSingleRequestHTTPServer(threading.Thread):
    def __init__(self, address, request_handler):
        threading.Thread.__init__(self, **{'group': None})
        self.daemon = True
        self.sucess = False
        self.shutdown = False
        self.httpd = HTTPServer(address, request_handler)
        self.httpd.timeout = 1.0
        self.assertions = self.httpd.assertions = []

    def run(self):
        self.httpd.handle_request()
        if not self.assertions:
            self.sucess = True
        # force socket close so next test can bind to same address
        self.httpd.socket.close()


def mock_http_handler(requests_responses, unordered=False, query_comparator=None):
    if query_comparator is None:
        query_comparator = query_eq

    class MockHTTPHandler(BaseHTTPRequestHandler):
        def do_HEAD(self):
            self.query_data = self.path
            return self.do_mock_request('HEAD')

        def do_GET(self):
            self.query_data = self.path
            return self.do_mock_request('GET')

        def do_POST(self):
            length = int(self.headers['content-length'])
            self.query_data = self.path + '?' + self.rfile.read(length).decode('utf-8')
            return self.do_mock_request('POST')

        def _matching_req_resp(self):
            if len(requests_responses) == 0:
                return None, None
            if unordered:
                for req_resp in requests_responses:
                    req, resp = req_resp
                    if query_comparator(req['path'], self.query_data):
                        requests_responses.remove(req_resp)
                        return req, resp
                return None, None
            else:
                return requests_responses.pop(0)

        def do_mock_request(self, method):
            req, resp = self._matching_req_resp()
            if not req:
                self.server.assertions.append(
                    RequestError('got unexpected request: %s' % self.query_data)
                )
                return
            if 'method' in req:
                if req['method'] != method:
                    self.server.assertions.append(
                        RequestMismatch('unexpected method', req['method'], method)
                    )
                    self.server.shutdown = True
            if req.get('require_basic_auth', False):
                if 'Authorization' not in self.headers:
                    requests_responses.insert(0, (req, resp))  # push back
                    self.send_response(401)
                    self.send_header('WWW-Authenticate', 'Basic realm="Secure Area"')
                    self.end_headers()
                    self.wfile.write(b'no access')
                    return
            if req.get('headers'):
                for k, v in req['headers'].items():
                    if k not in self.headers:
                        self.server.assertions.append(
                            RequestMismatch('missing header', k, self.headers)
                        )
                    elif self.headers[k] != v:
                        self.server.assertions.append(
                            RequestMismatch('header mismatch', '%s: %s' % (k, v), self.headers)
                        )
            if not query_comparator(req['path'], self.query_data):
                self.server.assertions.append(
                    RequestMismatch('requests differ', req['path'], self.query_data)
                )
                query_actual = set(query_to_dict(self.query_data).items())
                query_expected = set(query_to_dict(req['path']).items())
                self.server.assertions.append(
                    RequestMismatch('requests params differ', query_expected -
                                    query_actual, query_actual - query_expected)
                )
                self.server.shutdown = True
            if 'req_assert_function' in req:
                if not req['req_assert_function'](self):
                    self.server.assertions.append(
                        RequestError('req_assert_function failed')
                    )
                    self.server.shutdown = True
            if 'duration' in resp:
                time.sleep(float(resp['duration']))
            self.start_response(resp)
            if resp.get('body_file'):
                with open(resp['body_file'], 'rb') as f:
                    self.wfile.write(f.read())
            else:
                if 'body' in resp:
                    self.wfile.write(resp['body'])
                else:
                    self.wfile.write(b'')
            if not requests_responses:
                self.server.shutdown = True
            return

        def start_response(self, resp):
            self.send_response(int(resp.get('status', '200')))
            if 'headers' in resp:
                for key, value in resp['headers'].items():
                    self.send_header(key, value)
            self.end_headers()

        def log_request(self, code, size=None):
            pass

    return MockHTTPHandler


class MockServ(object):
    def __init__(self, port=0, host='localhost', unordered=False, bbox_aware_query_comparator=False):
        self._requested_port = port
        self.port = port
        self.host = host
        self.requests_responses = []
        self.unordered = unordered
        self.query_comparator = None
        if bbox_aware_query_comparator:
            self.query_comparator = wms_query_eq
        self._init_thread()

    def _init_thread(self):
        self._thread = ThreadedStopableHTTPServer((self.host, self._requested_port),
                                                  [], unordered=self.unordered, query_comparator=self.query_comparator)
        if self._requested_port == 0:
            self.port = self._thread.http_port
        self.address = (self.host, self.port)

    def reset(self):
        self._init_thread()

    @property
    def base_url(self):
        return 'http://localhost:%d' % (self.port, )

    def expects(self, path, method='GET', headers=None):
        headers = headers or ()
        self.requests_responses.append(
            (dict(path=path, method=method, headers=headers), {'body': b''}))
        return self

    def returns(self, body=None, body_file=None, status_code=200, headers=None):
        assert body or body_file
        headers = headers or {}
        self.requests_responses[-1][1].update(
            body=body, body_file=body_file, status=status_code, headers=headers)
        return self

    def __enter__(self):
        # copy request_responses to be able to reuse it after .reset()
        self._thread.requests_responses[:] = self.requests_responses
        self._thread.start()

    def __exit__(self, type, value, traceback):
        self._thread.shutdown = True
        self._thread.join(30)

        if not self._thread.sucess and value:
            print('requests to mock httpd did not '
                  'match expectations:\n %s' % RequestsMismatchError(self._thread.assertions))
        if value:
            raise reraise((type, value, traceback))
        if not self._thread.sucess:
            raise RequestsMismatchError(self._thread.assertions)


def wms_query_eq(expected, actual):
    """
    >>> wms_query_eq('bAR=baz&foo=bizz&bbOX=0,0,100000,100000', 'foO=bizz&BBOx=-.0001,0.01,99999.99,100000.09&bar=baz')
    True
    >>> wms_query_eq('bAR=baz&foo=bizz&bbOX=0,0,100000,100000', 'foO=bizz&BBOx=-.0001,0.01,99999.99,100000.11&bar=baz')
    False
    >>> wms_query_eq('/service?bar=baz&fOO=bizz', 'foo=bizz&bar=baz')
    False
    >>> wms_query_eq('/1/2/3.png', '/1/2/3.png')
    True
    >>> wms_query_eq('/1/2/3.png', '/1/2/0.png')
    False
    """
    from mapproxy.util.bbox import bbox_equals
    if path_from_query(expected) != path_from_query(actual):
        return False

    expected = query_to_dict(expected)
    actual = query_to_dict(actual)

    if 'bbox' in expected and 'bbox' in actual:
        expected = expected.copy()
        expected_bbox = [float(x) for x in expected.pop('bbox').split(',')]
        actual = actual.copy()
        actual_bbox = [float(x) for x in actual.pop('bbox').split(',')]
        if expected != actual:
            return False
        if not bbox_equals(expected_bbox, actual_bbox):
            return False
    else:
        if expected != actual:
            return False

    return True


numbers_only = re.compile(r'^-?\d+\.\d+(,-?\d+\.\d+)*$')


def query_eq(expected, actual):
    """
    >>> query_eq('bAR=baz&foo=bizz', 'foO=bizz&bar=baz')
    True
    >>> query_eq('/service?bar=baz&fOO=bizz', 'foo=bizz&bar=baz')
    False
    >>> query_eq('/1/2/3.png', '/1/2/3.png')
    True
    >>> query_eq('/1/2/3.png', '/1/2/0.png')
    False
    >>> query_eq('/map?point=2.9999999999,1.00000000001', '/map?point=3.0,1.0')
    True
    """

    if path_from_query(expected) != path_from_query(actual):
        return False

    expected = query_to_dict(expected)
    actual = query_to_dict(actual)

    if set(expected.keys()) != set(actual.keys()):
        return False

    for ke, ve in expected.items():
        if numbers_only.match(ve):
            if not float_string_almost_eq(ve, actual[ke]):
                return False
        else:
            if ve != actual[ke]:
                return False

    return True


def float_string_almost_eq(expected, actual):
    """
    Compares if two strings with comma-separated floats are almost equal.
    Strings must contain floats.

    >>> float_string_almost_eq('12345678900', '12345678901')
    False
    >>> float_string_almost_eq('12345678900.0', '12345678901.0')
    True

    >>> float_string_almost_eq('12345678900.0,-3.0', '12345678901.0,-2.9999999999')
    True
    """
    if not numbers_only.match(expected) or not numbers_only.match(actual):
        return False

    expected_nums = [float(x) for x in expected.split(',')]
    actual_nums = [float(x) for x in actual.split(',')]

    if len(expected_nums) != len(actual_nums):
        return False

    for e, a in zip(expected_nums, actual_nums):
        if abs(e - a) > abs((e+a)/2)/10e9:
            return False

    return True


def assert_query_eq(expected, actual, fuzzy_number_compare=False):
    path_actual = path_from_query(actual)
    path_expected = path_from_query(expected)
    assert path_expected == path_actual, path_expected + '!=' + path_actual

    query_actual = set(query_to_dict(actual).items())
    query_expected = set(query_to_dict(expected).items())

    if fuzzy_number_compare:
        equal = query_eq(expected, actual)
    else:
        equal = query_expected == query_actual
    assert equal, '%s != %s\t%s|%s' % (
        expected, actual, query_expected - query_actual, query_actual - query_expected)


def path_from_query(query):
    """
    >>> path_from_query('/service?foo=bar')
    '/service'
    >>> path_from_query('/1/2/3.png')
    '/1/2/3.png'
    >>> path_from_query('foo=bar')
    ''
    """
    if not ('&' in query or '=' in query):
        return query
    if '?' in query:
        return query.split('?', 1)[0]
    return ''


def query_to_dict(query):
    """
    >>> sorted(query_to_dict('/service?bar=baz&foo=bizz').items())
    [('bar', 'baz'), ('foo', 'bizz')]
    >>> sorted(query_to_dict('bar=baz&foo=bizz').items())
    [('bar', 'baz'), ('foo', 'bizz')]
    """
    if not ('&' in query or '=' in query):
        return {}
    d = {}
    if '?' in query:
        query = query.split('?', 1)[-1]
    for key, value in parse_qsl(query):
        d[key.lower()] = value
    return d


def assert_url_eq(url1, url2):
    parts1 = urlsplit(url1)
    parts2 = urlsplit(url2)

    assert parts1[0] == parts2[0], '%s != %s (%s)' % (url1, url2, 'schema')
    assert parts1[1] == parts2[1], '%s != %s (%s)' % (url1, url2, 'location')
    assert parts1[2] == parts2[2], '%s != %s (%s)' % (url1, url2, 'path')
    assert query_eq(parts1[3], parts2[3]), '%s != %s (%s)' % (url1, url2, 'query')
    assert parts1[4] == parts2[4], '%s != %s (%s)' % (url1, url2, 'fragment')


@contextmanager
def mock_httpd(address, requests_responses, unordered=False, bbox_aware_query_comparator=False):
    if bbox_aware_query_comparator:
        query_comparator = wms_query_eq
    else:
        query_comparator = query_eq
    t = ThreadedStopableHTTPServer(address, requests_responses, unordered=unordered,
                                   query_comparator=query_comparator)
    t.start()
    try:
        yield
    except Exception:
        if not t.sucess:
            print(str(RequestsMismatchError(t.assertions)))
        raise
    finally:
        t.shutdown = True
        t.join(30)
    if not t.sucess:
        raise RequestsMismatchError(t.assertions)


@contextmanager
def mock_single_req_httpd(address, request_handler):
    t = ThreadedSingleRequestHTTPServer(address, request_handler)
    t.start()
    try:
        yield
    except Exception:
        if not t.sucess:
            print(str(RequestsMismatchError(t.assertions)))
        raise
    finally:
        t.shutdown = True
        t.join(30)
    if not t.sucess:
        raise RequestsMismatchError(t.assertions)


def make_wsgi_env(query_string, extra_environ={}):
    env = {'QUERY_STRING': query_string,
           'wsgi.url_scheme': 'http',
           'HTTP_HOST': 'localhost',
           }
    env.update(extra_environ)
    return env


def basic_auth_value(username, password):
    return base64.b64encode(('%s:%s' % (username, password)).encode('utf-8'))


def assert_no_cache(resp):
    assert resp.headers["Pragma"] == "no-cache"
    assert resp.headers["Expires"] == "-1"
    assert resp.cache_control.no_store is True



# This file is part of the MapProxy project.
# Copyright (C) 2010 Omniscale <http://omniscale.de>
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import division

from io import BytesIO

import pytest

from mapproxy.request.wms import WMS111MapRequest
from PIL import Image
from mapproxy.test.image import is_png, tmp_image
from mapproxy.test.http import mock_httpd
from mapproxy.test.system import SysTest


@pytest.fixture(scope="module")
def config_file():
    return "coverage.yaml"


class TestCoverageWMS(SysTest):

    def setup_method(self):
        self.common_map_req = WMS111MapRequest(
            url="/service?",
            param=dict(
                service="WMS",
                version="1.1.1",
                bbox="-180,0,0,80",
                width="200",
                height="200",
                layers="wms_cache",
                srs="EPSG:4326",
                format="image/png",
                styles="",
                request="GetMap",
            ),
        )

    def test_capababilities(self, app):
        resp = app.get("/service?request=GetCapabilities&service=WMS&version=1.1.1")
        xml = resp.lxml
        # First: combined root, second: wms_cache, third: tms_cache, last: seed_only
        assert xml.xpath("//LatLonBoundingBox/@minx") == ["10", "10", "12", "14"]
        assert xml.xpath("//LatLonBoundingBox/@miny") == ["10", "15", "10", "13"]
        assert xml.xpath("//LatLonBoundingBox/@maxx") == ["35", "30", "35", "24"]
        assert xml.xpath("//LatLonBoundingBox/@maxy") == ["31", "31", "30", "23"]

    def test_get_map_outside(self, app):
        self.common_map_req.params.bbox = -90, 0, 0, 90
        self.common_map_req.params["bgcolor"] = "0xff0005"
        resp = app.get(self.common_map_req)
        assert resp.content_type == "image/png"
        data = BytesIO(resp.body)
        assert is_png(data)
        img = Image.open(data)
        assert img.mode == "RGB"
        assert img.getcolors() == [(200 * 200, (255, 0, 5))]

    def test_get_map_outside_transparent(self, app):
        self.common_map_req.params.bbox = -90, 0, 0, 90
        self.common_map_req.params.transparent = True
        resp = app.get(self.common_map_req)
        assert resp.content_type == "image/png"
        data = BytesIO(resp.body)
        assert is_png(data)
        img = Image.open(data)
        assert img.mode == "RGBA"
        assert img.getcolors()[0][0] == 200 * 200
        assert img.getcolors()[0][1][3] == 0  # transparent

    def test_get_map_intersection(self, app, cache_dir):
        with tmp_image((256, 256), format="jpeg") as img:
            expected_req = (
                {
                    "path": r"/service?LAYERs=foo,bar&SERVICE=WMS&FORMAT=image%2Fjpeg"
                    "&REQUEST=GetMap&HEIGHT=91&SRS=EPSG%3A4326&styles="
                    "&VERSION=1.1.1&BBOX=10,15,30,31"
                    "&WIDTH=114"
                },
                {"body": img.read(), "headers": {"content-type": "image/jpeg"}},
            )
            with mock_httpd(("localhost", 42423), [expected_req]):
                self.common_map_req.params.bbox = 0, 0, 40, 40
                self.common_map_req.params.transparent = True
                resp = app.get(self.common_map_req)
                assert resp.content_type == "image/png"
                data = BytesIO(resp.body)
                assert is_png(data)
                assert Image.open(data).mode == "RGBA"
        assert cache_dir.join(
            "wms_cache_EPSG4326/03/000/000/004/000/000/002.jpeg"
        ).check()


class TestCoverageTMS(SysTest):

    def test_get_tile_intersections(self, app, cache_dir):
        with tmp_image((256, 256), format="jpeg") as img:
            expected_req = (
                {
                    "path": r"/service?LAYERs=foo,bar&SERVICE=WMS&FORMAT=image%2Fjpeg"
                    "&REQUEST=GetMap&HEIGHT=25&SRS=EPSG%3A900913&styles="
                    "&VERSION=1.1.1&BBOX=1113194.90793,1689200.13961,3339584.7238,3632749.14338"
                    "&WIDTH=28"
                },
                {"body": img.read(), "headers": {"content-type": "image/jpeg"}},
            )
            with mock_httpd(
                ("localhost", 42423), [expected_req], bbox_aware_query_comparator=True
            ):
                resp = app.get("/tms/1.0.0/wms_cache/0/1/1.jpeg")
                assert resp.content_type == "image/jpeg"
        cache_dir.join("wms_cache_EPSG900913/01/000/000/001/000/000/001.jpeg").check()

    def test_get_tile_intersection_tms(self, app, cache_dir):
        with tmp_image((256, 256), format="jpeg") as img:
            expected_req = (
                {"path": r"/tms/1.0.0/foo/1/1/1.jpeg"},
                {"body": img.read(), "headers": {"content-type": "image/jpeg"}},
            )
            with mock_httpd(
                ("localhost", 42423), [expected_req], bbox_aware_query_comparator=True
            ):
                resp = app.get("/tms/1.0.0/tms_cache/0/1/1.jpeg")
                assert resp.content_type == "image/jpeg"
        cache_dir.join("tms_cache_EPSG900913/01/000/000/001/000/000/001.jpeg").check()



# This file is part of the MapProxy project.
# Copyright (C) 2012 Omniscale <http://omniscale.de>
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from PIL import Image, ImageDraw
from mapproxy.image import ImageSource
from mapproxy.image.mask import mask_image_source_from_coverage
from mapproxy.image.merge import LayerMerger
from mapproxy.image.opts import ImageOptions
from mapproxy.srs import SRS
from mapproxy.test.image import assert_img_colors_eq, create_image
from mapproxy.util.coverage import load_limited_to

from shapely.geometry import Polygon


def coverage(geom, srs="EPSG:4326"):
    return load_limited_to({"srs": srs, "geometry": geom})


class TestMaskImage(object):

    def test_mask_outside_of_image_transparent(self):
        img = ImageSource(
            Image.new("RGB", (100, 100), color=(100, 0, 200)),
            image_opts=ImageOptions(transparent=True),
        )
        result = mask_image_source_from_coverage(
            img, [0, 0, 10, 10], SRS(4326), coverage([20, 20, 30, 30])
        )
        assert_img_colors_eq(
            result.as_image().getcolors(), [((100 * 100), (255, 255, 255, 0))]
        )

    def test_mask_outside_of_image_bgcolor(self):
        img = ImageSource(
            Image.new("RGB", (100, 100), color=(100, 0, 200)),
            image_opts=ImageOptions(bgcolor=(200, 30, 120)),
        )

        result = mask_image_source_from_coverage(
            img, [0, 0, 10, 10], SRS(4326), coverage([20, 20, 30, 30])
        )
        assert_img_colors_eq(
            result.as_image().getcolors(), [((100 * 100), (200, 30, 120))]
        )

    def test_mask_partial_image_bgcolor(self):
        img = ImageSource(
            Image.new("RGB", (100, 100), color=(100, 0, 200)),
            image_opts=ImageOptions(bgcolor=(200, 30, 120)),
        )

        result = mask_image_source_from_coverage(
            img, [0, 0, 10, 10], SRS(4326), coverage([5, 5, 30, 30])
        )
        assert_img_colors_eq(
            result.as_image().getcolors(),
            [(7500, (200, 30, 120)), (2500, (100, 0, 200))],
        )

    def test_mask_partial_image_transparent(self):
        img = ImageSource(
            Image.new("RGB", (100, 100), color=(100, 0, 200)),
            image_opts=ImageOptions(transparent=True),
        )

        result = mask_image_source_from_coverage(
            img, [0, 0, 10, 10], SRS(4326), coverage([5, 5, 30, 30])
        )
        assert_img_colors_eq(
            result.as_image().getcolors(),
            [(7500, (255, 255, 255, 0)), (2500, (100, 0, 200, 255))],
        )

    def test_wkt_mask_partial_image_transparent(self):
        img = ImageSource(
            Image.new("RGB", (100, 100), color=(100, 0, 200)),
            image_opts=ImageOptions(transparent=True),
        )

        # polygon with hole
        geom = "POLYGON((2 2, 2 8, 8 8, 8 2, 2 2), (4 4, 4 6, 6 6, 6 4, 4 4))"

        result = mask_image_source_from_coverage(
            img, [0, 0, 10, 10], SRS(4326), coverage(geom)
        )
        # 60*60 - 20*20 = 3200
        assert_img_colors_eq(
            result.as_image().getcolors(),
            [(10000 - 3200, (255, 255, 255, 0)), (3200, (100, 0, 200, 255))],
        )

    def test_shapely_mask_with_transform_partial_image_transparent(self):
        img = ImageSource(
            Image.new("RGB", (100, 100), color=(100, 0, 200)),
            image_opts=ImageOptions(transparent=True),
        )

        p = Polygon(
            [(0, 0), (222000, 0), (222000, 222000), (0, 222000)]
        )  # ~ 2x2 degres

        result = mask_image_source_from_coverage(
            img, [0, 0, 10, 10], SRS(4326), coverage(p, "EPSG:3857")
        )
        # 20*20 = 400
        assert_img_colors_eq(
            result.as_image().getcolors(),
            [(10000 - 400, (255, 255, 255, 0)), (400, (100, 0, 200, 255))],
        )


class TestLayerCoverageMerge(object):

    def setup_method(self):
        self.coverage1 = coverage(
            Polygon([(0, 0), (0, 10), (10, 10), (10, 0)]), 3857
        )
        self.coverage2 = coverage([2, 2, 8, 8], 3857)

    def test_merge_single_coverage(self):
        merger = LayerMerger()
        merger.add(
            ImageSource(Image.new("RGB", (10, 10), (255, 255, 255))),
            self.coverage1,
        )
        result = merger.merge(
            image_opts=ImageOptions(transparent=True),
            bbox=(5, 0, 15, 10),
            bbox_srs=3857,
        )
        img = result.as_image()
        assert img.mode == "RGBA"
        assert img.getpixel((4, 0)) == (255, 255, 255, 255)
        assert img.getpixel((6, 0)) == (255, 255, 255, 0)

    def test_merge_overlapping_coverage(self):
        color1 = (255, 255, 0)
        color2 = (0, 255, 255)
        merger = LayerMerger()
        merger.add(
            ImageSource(Image.new("RGB", (10, 10), color1)), self.coverage1
        )
        merger.add(
            ImageSource(Image.new("RGB", (10, 10), color2)), self.coverage2
        )

        result = merger.merge(
            image_opts=ImageOptions(), bbox=(0, 0, 10, 10), bbox_srs=3857
        )
        img = result.as_image()
        assert img.mode == "RGB"

        expected = create_image((10, 10), color1, "RGB")
        draw = ImageDraw.Draw(expected)
        draw.polygon([(2, 2), (7, 2), (7, 7), (2, 7)], fill=color2)

        for x in range(0, 9):
            for y in range(0, 9):
                assert img.getpixel((x, y)) == expected.getpixel((x, y))



# This file is part of the MapProxy project.
# Copyright (C) 2011 Omniscale <http://omniscale.de>
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import print_function

import time
import threading

from mapproxy.util.async_ import ThreadPool, imap


class TestThreaded(object):
    def test_map(self):
        def func(x):
            time.sleep(0.05)
            return x
        start = time.time()
        result = list(imap(func, list(range(40))))
        stop = time.time()

        duration = stop - start
        assert duration < 0.5, "took %s" % duration

        assert len(result) == 40

    def test_map_with_exception(self):
        def func(x):
            raise Exception()

        try:
            list(imap(func, list(range(40))))
        except Exception:
            pass
        else:
            assert False, 'exception expected'


class CommonPoolTests(object):
    def _check_single_arg(self, func):
        result = list(func())
        assert result == [3]

    def test_single_argument(self):
        def f1(x, y): return x+y
        pool = self.mk_pool()
        check = self._check_single_arg
        check(lambda: pool.map(f1, [1], [2]))
        check(lambda: pool.imap(f1, [1], [2]))
        check(lambda: pool.starmap(f1, [(1, 2)]))
        check(lambda: pool.starcall([(f1, 1, 2)]))

    def _check_single_arg_raise(self, func):
        try:
            list(func())
        except ValueError:
            pass
        else:
            assert False, 'expected ValueError'

    def test_single_argument_raise(self):
        def f1(x, y):
            raise ValueError
        pool = self.mk_pool()
        check = self._check_single_arg_raise
        check(lambda: pool.map(f1, [1], [2]))
        check(lambda: pool.imap(f1, [1], [2]))
        check(lambda: pool.starmap(f1, [(1, 2)]))
        check(lambda: pool.starcall([(f1, 1, 2)]))

    def _check_single_arg_result_object(self, func):
        result = list(func())
        assert result[0].result is None
        assert isinstance(result[0].exception[1], ValueError)

    def test_single_argument_result_object(self):
        def f1(x, y):
            raise ValueError
        pool = self.mk_pool()
        check = self._check_single_arg_result_object
        check(lambda: pool.map(f1, [1], [2], use_result_objects=True))
        check(lambda: pool.imap(f1, [1], [2], use_result_objects=True))
        check(lambda: pool.starmap(f1, [(1, 2)], use_result_objects=True))
        check(lambda: pool.starcall([(f1, 1, 2)], use_result_objects=True))

    def _check_multiple_args(self, func):
        result = list(func())
        assert result == [3, 5]

    def test_multiple_arguments(self):
        def f1(x, y): return x+y
        pool = self.mk_pool()
        check = self._check_multiple_args
        check(lambda: pool.map(f1, [1, 2], [2, 3]))
        check(lambda: pool.imap(f1, [1, 2], [2, 3]))
        check(lambda: pool.starmap(f1, [(1, 2), (2, 3)]))
        check(lambda: pool.starcall([(f1, 1, 2), (f1, 2, 3)]))

    def _check_multiple_args_with_exceptions_result_object(self, func):
        result = list(func())
        assert result[0].result == 3
        assert isinstance(result[1].exception[1], ValueError)
        assert result[2].result == 7

    def test_multiple_arguments_exceptions_result_object(self):
        def f1(x, y):
            if x+y == 5:
                raise ValueError()
            return x+y
        pool = self.mk_pool()
        check = self._check_multiple_args_with_exceptions_result_object
        check(lambda: pool.map(f1, [1, 2, 3], [2, 3, 4], use_result_objects=True))
        check(lambda: pool.imap(f1, [1, 2, 3], [2, 3, 4], use_result_objects=True))
        check(lambda: pool.starmap(f1, [(1, 2), (2, 3), (3, 4)], use_result_objects=True))
        check(lambda: pool.starcall([(f1, 1, 2), (f1, 2, 3), (f1, 3, 4)], use_result_objects=True))

    def _check_multiple_args_with_exceptions(self, func):
        result = func()
        try:
            # first result might aleady raise the exception when
            # when second result is returned faster by the ThreadPoolWorker
            assert next(result) == 3
            next(result)
        except ValueError:
            pass
        else:
            assert False, 'expected ValueError'

    def test_multiple_arguments_exceptions(self):
        def f1(x, y):
            if x+y == 5:
                raise ValueError()
            return x+y
        pool = self.mk_pool()
        check = self._check_multiple_args_with_exceptions

        def check_pool_map():
            try:
                pool.map(f1, [1, 2, 3], [2, 3, 4])
            except ValueError:
                pass
            else:
                assert False, 'expected ValueError'
        check_pool_map()
        check(lambda: pool.imap(f1, [1, 2, 3], [2, 3, 4]))
        check(lambda: pool.starmap(f1, [(1, 2), (2, 3), (3, 4)]))
        check(lambda: pool.starcall([(f1, 1, 2), (f1, 2, 3), (f1, 3, 4)]))


class TestThreadPool(CommonPoolTests):
    def mk_pool(self):
        return ThreadPool()

    def test_base_config(self):
        # test that all concurrent have access to their
        # local base_config
        from mapproxy.config import base_config
        from mapproxy.config import local_base_config
        from copy import deepcopy

        # make two separate base_configs
        conf1 = deepcopy(base_config())
        conf1.conf = 1
        conf2 = deepcopy(base_config())
        conf2.conf = 2
        base_config().bar = 'baz'

        # run test in parallel, check1 and check2 should interleave
        # each with their local conf

        error_occured = False

        def check1(x):
            global error_occured
            if base_config().conf != 1 or 'bar' in base_config():
                error_occured = True

        def check2(x):
            global error_occured
            if base_config().conf != 2 or 'bar' in base_config():
                error_occured = True

        assert 'bar' in base_config()

        def test1():
            with local_base_config(conf1):
                pool1 = ThreadPool(5)
                list(pool1.imap(check1, list(range(200))))

        def test2():
            with local_base_config(conf2):
                pool2 = ThreadPool(5)
                list(pool2.imap(check2, list(range(200))))

        t1 = threading.Thread(target=test1)
        t2 = threading.Thread(target=test2)
        t1.start()
        t2.start()
        t1.join()
        t2.join()
        assert not error_occured
        assert 'bar' in base_config()


class DummyException(Exception):
    pass


class TestThreadedExecutorException(object):
    def setup_method(self):
        self.lock = threading.Lock()
        self.exec_count = 0
        self.te = ThreadPool(size=2)

    def execute(self, x):
        time.sleep(0.005)
        with self.lock:
            self.exec_count += 1
            if self.exec_count == 7:
                raise DummyException()
        return x

    def test_execute_w_exception(self):
        try:
            self.te.map(self.execute, list(range(100)))
        except DummyException:
            print(self.exec_count)
            assert 7 <= self.exec_count <= 10, 'execution should be interrupted really '\
                                               'soon (exec_count should be 7+(max(3)))'
        else:
            assert False, 'expected DummyException'



# This file is part of the MapProxy project.
# Copyright (C) 2010 Omniscale <http://omniscale.de>
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import division

import os
import shutil

import pytest

from webtest import TestApp as _TestApp

from mapproxy.wsgiapp import make_wsgi_app


class WSGITestApp(_TestApp):
    """
    Wraps webtest.TestApp and explicitly converts URLs to strings.
    Behavior changed with webtest from 1.2->1.3.
    """

    def get(self, url, *args, **kw):
        return _TestApp.get(self, str(url), *args, **kw)

    def head(self, url, *args, **kw):
        return _TestApp.head(self, str(url), *args, **kw)


@pytest.mark.usefixtures("cache_dir")
class SysTest(object):
    """
    Baseclass for pytest-based system tests.
    Provides `app` fixture with a configured MapProxy instance, wrapped in
    webtest.TestApp.

    `app` is reused within each test class, `cache_dir` is cleaned with
    each test.
    """

    @pytest.fixture(scope="class")
    def app(self, base_dir, config_file, additional_files):
        filename = base_dir.join(config_file)
        app = make_wsgi_app(filename.strpath, ignore_config_warnings=False)
        app.base_config.debug_mode = True
        return WSGITestApp(app, use_unicode=False)

    @pytest.fixture(scope="class")
    def additional_files(self, base_dir):
        return

    @pytest.fixture(scope="class")
    def base_config(self, app):
        return app.app.base_config

    @pytest.fixture(scope="class")
    def base_dir(self, tmpdir_factory, config_file):
        dir = tmpdir_factory.mktemp("base_dir")

        fixture_dir = os.path.join(os.path.dirname(__file__), "fixture")
        fixture_layer_conf = os.path.join(fixture_dir, config_file)
        shutil.copy(fixture_layer_conf, dir.strpath)

        return dir

    @pytest.fixture(scope="function")
    def cache_dir(self, base_dir):
        """
        cache_dir fixture returns a fresh cache_data directory used by `app`.
        """
        cache_dir = base_dir.join("cache_data")
        if cache_dir.check():
            cache_dir.remove()

        yield cache_dir

        if cache_dir.check():
            cache_dir.remove()

    @pytest.fixture(scope="function")
    def fixture_cache_data(self, cache_dir):
        """
        fixture_cache_data ensures that the system/fixture files are copied into
        `cache_dir`.
        """
        fixture_dir = os.path.join(os.path.dirname(__file__), "fixture")
        shutil.copytree(os.path.join(fixture_dir, "cache_data"), cache_dir.strpath)



# This file is part of the MapProxy project.
# Copyright (C) 2010 Omniscale <http://omniscale.de>
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
ctypes based replacement of pyroj (with pyproj fallback) of the old PROJ 4 API.

This module implements the `Proj`, `transform` and `set_datapath` class/functions. This
module is a drop-in replacement for pyproj. It does implement just enough to work for
MapProxy, i.e. there is no numpy support, etc.

It uses the C libproj library. If the library could not be found/loaded it will fallback
to pyroj. You can force the usage of either backend by setting the environment variables
MAPPROXY_USE_LIBPROJ or MAPPROXY_USE_PYPROJ to any value.

The new PROJ >=5 API is only supported via pyproj. See USE_PROJ4_API.

"""
from __future__ import print_function
import logging

import os
import sys
from mapproxy.util.lib import load_library

import ctypes
from ctypes import (
    c_void_p,
    c_char_p,
    c_int,
    c_double,
    c_long,
    POINTER,
    create_string_buffer,
    addressof,
)

c_double_p = POINTER(c_double)
FINDERCMD = ctypes.CFUNCTYPE(c_char_p, c_char_p)

log_system = logging.getLogger('mapproxy.system')

__all__ = ['Proj', 'transform', 'set_datapath', 'ProjError']


def init_libproj():
    libproj = load_library('libproj')

    if libproj is None:
        return

    if hasattr(libproj, 'proj_create'):
        log_system.warning('Found libproj >=5. Using this library without pyproj is '
                           'deprecated and not fully supported. Please install pyproj >= 2.')

    libproj.pj_init_plus.argtypes = [c_char_p]
    libproj.pj_init_plus.restype = c_void_p

    libproj.pj_is_latlong.argtypes = [c_void_p]
    libproj.pj_is_latlong.restype = c_int

    libproj.pj_get_def.argtypes = [c_void_p, c_int]
    libproj.pj_get_def.restype = c_void_p

    libproj.pj_strerrno.argtypes = [c_int]
    libproj.pj_strerrno.restype = c_char_p

    libproj.pj_get_errno_ref.argtypes = []
    libproj.pj_get_errno_ref.restype = POINTER(c_int)

    # free proj objects
    libproj.pj_free.argtypes = [c_void_p]
    # free() wrapper
    libproj.pj_dalloc.argtypes = [c_void_p]

    libproj.pj_transform.argtypes = [c_void_p, c_void_p, c_long, c_int,
                                     c_double_p, c_double_p, c_double_p]
    libproj.pj_transform.restype = c_int

    if hasattr(libproj, 'pj_set_searchpath'):
        libproj.pj_set_searchpath.argtypes = [c_int, POINTER(c_char_p)]
        libproj.pj_set_finder.argtypes = [FINDERCMD]

    return libproj


class SearchPath(object):
    def __init__(self):
        self.path = None
        self.finder_results = {}

    def clear(self):
        self.path = None
        self.finder_results = {}

    def set_searchpath(self, path):
        self.clear()
        if path is not None:
            path = path.encode(sys.getfilesystemencoding() or 'utf-8')
        self.path = path

    def finder(self, name):
        if self.path is None:
            return None

        if name in self.finder_results:
            result = self.finder_results[name]
        else:
            sysname = os.path.join(self.path, name)
            result = self.finder_results[name] = create_string_buffer(sysname)

        return addressof(result)


# search_path and finder_func must be defined in module
# context to avoid garbage collection
search_path = SearchPath()
finder_func = FINDERCMD(search_path.finder)
_finder_callback_set = False


class ProjError(RuntimeError):
    pass


class ProjInitError(ProjError):
    pass


def try_pyproj4_import():
    try:
        from pyproj import Proj, transform, set_datapath
    except ImportError:
        return False
    log_system.info('using pyproj with old Proj4 API for coordinate transformation')
    return Proj, transform, set_datapath


def try_pyproj_import():
    try:
        from pyproj import CRS
        from pyproj.transformer import Transformer
        from pyproj.datadir import set_data_dir
    except ImportError:
        return False
    log_system.info('using pyproj for coordinate transformation')
    return CRS, Transformer, set_data_dir


def try_libproj_import():
    libproj = init_libproj()

    if libproj is None:
        return False

    log_system.info('using libproj for coordinate transformation')

    RAD_TO_DEG = 57.29577951308232
    DEG_TO_RAD = .0174532925199432958

    class Proj(object):
        def __init__(self, proj_def=None, init=None):
            if init:
                self._proj = libproj.pj_init_plus(b'+init=' + init.encode('ascii'))
            else:
                self._proj = libproj.pj_init_plus(proj_def.encode('ascii'))
            if not self._proj:
                errno = libproj.pj_get_errno_ref().contents
                raise ProjInitError('error initializing Proj(proj_def=%r, init=%r): %s' %
                                    (proj_def, init, libproj.pj_strerrno(errno)))

            self.srs = self._srs()
            self._latlong = bool(libproj.pj_is_latlong(self._proj))

        def is_latlong(self):
            """
            >>> Proj(init='epsg:4326').is_latlong()
            True
            >>> Proj(init='epsg:4258').is_latlong()
            True
            >>> Proj(init='epsg:31467').is_latlong()
            False
            >>> Proj('+proj=merc +a=6378137 +b=6378137 +lat_ts=0.0 '
            ...      '+lon_0=0.0 +x_0=0.0 +y_0=0 +k=1.0 +units=m '
            ...      '+nadgrids=@null +no_defs').is_latlong()
            False
            """
            return self._latlong

        def _srs(self):
            res = libproj.pj_get_def(self._proj, 0)
            srs_def = ctypes.c_char_p(res).value
            libproj.pj_dalloc(res)
            return srs_def

        def __del__(self):
            if self._proj and libproj:
                libproj.pj_free(self._proj)
                self._proj = None

    def transform(from_srs, to_srs, x, y, z=None):
        if from_srs == to_srs:
            return (x, y) if z is None else (x, y, z)

        if isinstance(x, (float, int)):
            x = [x]
            y = [y]
        assert len(x) == len(y)

        if from_srs.is_latlong():
            x = [x*DEG_TO_RAD for x in x]
            y = [y*DEG_TO_RAD for y in y]

        x = (c_double * len(x))(*x)
        y = (c_double * len(y))(*y)
        if z is not None:
            z = (c_double * len(z))(*z)
        else:
            # use explicit null pointer instead of None
            # http://bugs.python.org/issue4606
            z = c_double_p()

        res = libproj.pj_transform(from_srs._proj, to_srs._proj,
                                   len(x), 0, x, y, z)

        if res:
            raise ProjError(libproj.pj_strerrno(res))

        if to_srs.is_latlong():
            x = [x*RAD_TO_DEG for x in x]
            y = [y*RAD_TO_DEG for y in y]
        else:
            x = x[:]
            y = y[:]

        if len(x) == 1:
            x = x[0]
            y = y[0]
            z = z[0] if z else None

        return (x, y) if z is None else (x, y, z)

    def set_datapath(path):
        global _finder_callback_set
        if not _finder_callback_set:
            libproj.pj_set_finder(finder_func)
            _finder_callback_set = True
        search_path.set_searchpath(path)

    return Proj, transform, set_datapath


proj_imports = []

if 'MAPPROXY_USE_LIBPROJ' in os.environ:
    proj_imports = [try_libproj_import]

if 'MAPPROXY_USE_PYPROJ' in os.environ:
    proj_imports = [try_pyproj_import, try_pyproj4_import]

if not proj_imports:
    if sys.platform == 'win32':
        # prefer pyproj on windows
        proj_imports = [try_pyproj_import, try_pyproj4_import, try_libproj_import]
    else:
        proj_imports = [try_pyproj_import, try_libproj_import, try_pyproj4_import]

# try different imports in previously defined order
for try_import in proj_imports:
    if try_import == try_pyproj_import:
        res = try_import()
        if res:
            CRS, Transformer, set_datapath = res
            Proj, transform = None, None
            USE_PROJ4_API = False
            break
    else:
        res = try_import()
        if res:
            Proj, transform, set_datapath = res
            CRS, Transformer = None, None
            USE_PROJ4_API = True
            break
else:
    raise ImportError('could not find pyproj (Python library) or libproj (C library, deprecated)')

if __name__ == '__main__':

    prj1 = Proj(init='epsg:4326')
    prj2 = Proj(init='epsg:31467')

    coords = [(8.2, 8.22, 8.3), (53.1, 53.15, 53.2)]
    # coords = [(8, 9, 10), (50, 50, 50)]
    print(coords)
    coords = transform(prj1, prj2, *coords)
    print(coords)
    coords = transform(prj2, prj1, *coords)
    print(coords)



# This file is part of the MapProxy project.
# Copyright (C) 2015 Omniscale <http://omniscale.de>
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import print_function

import yaml

from mapproxy.config.validator import validate


class TestValidator(object):
    def _test_conf(self, yaml_part=None):
        base = yaml.safe_load('''
            services:
                wms:
                    md:
                        title: MapProxy
            layers:
                - name: one
                  title: One
                  sources: [one_cache]
            caches:
                one_cache:
                    grids: [GLOBAL_MERCATOR]
                    sources: [one_source]
            sources:
                one_source:
                    type: wms
                    req:
                        url: http://localhost/service?
                        layers: one
        ''')
        if yaml_part is not None:
            base.update(yaml.safe_load(yaml_part))
        return base

    def test_valid_config(self):
        conf = self._test_conf()

        errors = validate(conf)
        assert errors == []

    def test_missing_layer_source(self):
        conf = self._test_conf()
        del conf['caches']['one_cache']

        errors = validate(conf)
        assert errors == [
            "Source 'one_cache' for layer 'one' not in cache or source section"
        ]

    def test_multiple_errors(self):
        conf = self._test_conf()
        del conf['caches']['one_cache']
        del conf['services']
        errors = validate(conf)
        assert errors == [
            "'services' is a required property in root",
            "Source 'one_cache' for layer 'one' not in cache or source section",
        ]

    def test_empty_layer_sources(self):
        conf = self._test_conf('''
            layers:
                - name: one
                  title: One
                  sources: []
        ''')

        errors = validate(conf)
        assert errors in [
            [
                "[] is too short in root.layers[0].sources"
            ],
            [
                "[] should be non-empty in root.layers[0].sources"
            ]
        ]

    def test_missing_cache_source(self):
        conf = self._test_conf()
        del conf['sources']['one_source']

        errors = validate(conf)
        assert errors == [
            "Source 'one_source' for cache 'one_cache' not found in config"
        ]

    def test_missing_layers_section(self):
        conf = self._test_conf()
        del conf['layers']

        errors = validate(conf)
        assert errors == [
            "'layers' is a required property in root"
        ]

    def test_missing_services_section(self):
        conf = self._test_conf()
        del conf['services']
        errors = validate(conf)
        assert errors == [
            "'services' is a required property in root"
        ]

    def test_tile_source(self):
        conf = self._test_conf('''
            layers:
                - name: one
                  tile_sources: [missing]
        ''')

        errors = validate(conf)
        assert errors == [
            "Tile source 'missing' for layer 'one' not in cache section"
        ]

    def test_missing_grid(self):
        conf = self._test_conf('''
            caches:
                one_cache:
                    grids: [MYGRID_OTHERGRID]
                    sources: []
            grids:
                MYGRID:
                    base: GLOBAL_GEODETIC
        ''')

        errors = validate(conf)
        assert errors == [
            "Grid 'MYGRID_OTHERGRID' for cache 'one_cache' not found in config"
        ]

    def test_misconfigured_wms_source(self):
        conf = self._test_conf()

        del conf['sources']['one_source']['req']['layers']

        errors = validate(conf)
        assert errors == [
            "Missing 'layers' for source 'one_source'"
        ]

    def test_misconfigured_mapserver_source_without_globals(self):
        conf = self._test_conf('''
            sources:
                one_source:
                    type: mapserver
                    req:
                        map: foo.map
                    mapserver:
                        binary: /foo/bar/baz
        ''')

        errors = validate(conf)
        assert errors == [
            'Could not find mapserver binary (/foo/bar/baz)'
        ]

        del conf['sources']['one_source']['mapserver']['binary']

        errors = validate(conf)
        assert errors == [
            "Missing mapserver binary for source 'one_source'"
        ]

        del conf['sources']['one_source']['mapserver']

        errors = validate(conf)
        assert errors == [
            "Missing mapserver binary for source 'one_source'"
        ]

    def test_misconfigured_mapserver_source_with_globals(self):
        conf = self._test_conf('''
            sources:
                one_source:
                    type: mapserver
                    req:
                        map: foo.map
            globals:
                mapserver:
                    binary: /foo/bar/baz
        ''')

        errors = validate(conf)
        assert errors == [
            'Could not find mapserver binary (/foo/bar/baz)'
        ]

        del conf['globals']['mapserver']['binary']

        errors = validate(conf)
        assert errors == [
            "Missing mapserver binary for source 'one_source'"
        ]

    def test_tagged_sources_with_layers(self):
        conf = self._test_conf('''
            caches:
                one_cache:
                    grids: [GLOBAL_MERCATOR]
                    sources: ['one_source:foo,bar']
        ''')

        errors = validate(conf)
        assert errors == [
            "Supported layers for source 'one_source' are 'one' but tagged source "
            "requested layers 'foo, bar'"
        ]

    def test_tagged_layer_sources_with_layers(self):
        conf = self._test_conf('''
            layers:
                - name: one
                  title: One
                  sources: ['one_source:foo,bar']
        ''')

        errors = validate(conf)
        assert errors == [
            "Supported layers for source 'one_source' are 'one' but tagged source "
            "requested layers 'foo, bar'"
        ]

    def test_tagged_layer_sources_without_layers(self):
        conf = self._test_conf('''
            layers:
                - name: one
                  title: One
                  sources: ['one_source:foo,bar']
        ''')

        del conf['sources']['one_source']['req']['layers']

        errors = validate(conf)
        assert errors == []

    def test_tagged_source_without_layers(self):
        conf = self._test_conf('''
            caches:
                one_cache:
                    grids: [GLOBAL_MERCATOR]
                    sources: ['one_source:foo,bar']
        ''')

        del conf['sources']['one_source']['req']['layers']

        errors = validate(conf)
        assert errors == []

    def test_tagged_source_with_colons(self):
        conf = self._test_conf('''
            caches:
                one_cache:
                    grids: [GLOBAL_MERCATOR]
                    sources: ['one_source:ns:foo,ns:bar']
        ''')

        del conf['sources']['one_source']['req']['layers']

        errors = validate(conf)
        assert errors == []

    def test_with_grouped_layer(self):
        conf = self._test_conf('''
            layers:
                - name: group
                  title: Group
                  layers:
                    - name: one
                      title: One
                      sources: [one_cache]
        ''')

        errors = validate(conf)
        assert errors == []

    def test_without_cache(self):
        conf = self._test_conf('''
            layers:
              - name: one
                title: One
                sources: [one_source]
        ''')

        errors = validate(conf)
        assert errors == []

    def test_mapserver_with_tagged_layers(self):
        conf = self._test_conf('''
            sources:
                one_source:
                    type: mapserver
                    req:
                        map: foo.map
                        layers: one
                    mapserver:
                        binary: /foo/bar/baz
            caches:
                one_cache:
                    grids: [GLOBAL_MERCATOR]
                    sources: ['one_source:foo,bar']
        ''')

        errors = validate(conf)
        assert errors == [
            'Could not find mapserver binary (/foo/bar/baz)',
            "Supported layers for source 'one_source' are 'one' but tagged source "
            "requested layers 'foo, bar'"
        ]

    def test_mapnik_with_tagged_layers(self):
        conf = self._test_conf('''
            sources:
                one_source:
                    type: mapnik
                    mapfile: foo.map
                    layers: one
            caches:
                one_cache:
                    grids: [GLOBAL_MERCATOR]
                    sources: ['one_source:foo,bar']
        ''')

        errors = validate(conf)
        assert errors == [
            "Supported layers for source 'one_source' are 'one' but tagged source "
            "requested layers 'foo, bar'"
        ]

    def test_tagged_layers_for_unsupported_source_type(self):
        conf = self._test_conf('''
            sources:
                one_source:
                    type: tile
                    url: http://localhost/tiles/
            caches:
                one_cache:
                    grids: [GLOBAL_MERCATOR]
                    sources: ['one_source:foo,bar']
        ''')

        errors = validate(conf)
        assert errors == [
            "Found tagged source 'one_source' in cache 'one_cache' but tagged sources "
            "only supported for 'wms, mapserver, mapnik' sources"
        ]

    def test_cascaded_caches(self):
        conf = self._test_conf('''
            caches:
                one_cache:
                    sources: [two_cache]
                two_cache:
                    grids: [GLOBAL_MERCATOR]
                    sources: ['one_source']
        ''')

        errors = validate(conf)
        assert errors == []

    def test_with_int_0_as_names_and_layers(self):
        conf = self._test_conf('''
            services:
                wms:
                    md:
                        title: MapProxy
            layers:
                - name: 0
                  title: One
                  sources: [0]
            caches:
                0:
                    grids: [GLOBAL_MERCATOR]
                    sources: [0]
            sources:
                0:
                    type: wms
                    req:
                        url: http://localhost/service?
                        layers: 0
        ''')

        errors = validate(conf)
        assert errors == []

    def test_band_merge_missing_source(self):
        conf = self._test_conf('''
            caches:
                one_cache:
                    sources:
                        l:
                            - source: dop
                              band: 1
                              factor: 0.4
                            - source: missing1
                              band: 2
                              factor: 0.2
                            - source: cache_missing_source
                              band: 2
                              factor: 0.2
                    grids: [GLOBAL_MERCATOR]
                cache_missing_source:
                    sources: [missing2]
                    grids: [GLOBAL_MERCATOR]

            sources:
                dop:
                    type: wms
                    req:
                        url: http://localhost/service?
                        layers: dop
        ''')

        errors = validate(conf)
        assert errors == [
            "Source 'missing1' for cache 'one_cache' not found in config",
            "Source 'missing2' for cache 'cache_missing_source' not found in config",
        ]



# This file is part of the MapProxy project.
# Copyright (C) 2014 Omniscale <http://omniscale.de>
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import division

from mapproxy.test.system import SysTest
from mapproxy.test.system.test_wms import is_110_capa, is_111_capa, is_130_capa

import pytest


class TestLimitedWMSVersionsTest(SysTest):

    @pytest.fixture(scope="class")
    def config_file(self):
        return "wms_versions.yaml"

    def test_default_version_130(self, app):
        resp = app.get("http://localhost/service?SERVICE=WMS&REQUEST=GetCapabilities")
        assert is_111_capa(resp.lxml)

    def test_supported_version_110(self, app):
        resp = app.get(
            "http://localhost/service?SERVICE=WMS&REQUEST=GetCapabilities"
            "&VERSION=1.1.0"
        )
        assert is_110_capa(resp.lxml)

    def test_unknown_version_113(self, app):
        resp = app.get(
            "http://localhost/service?SERVICE=WMS&REQUEST=GetCapabilities"
            "&VERSION=1.1.3"
        )
        assert is_111_capa(resp.lxml)

    def test_unknown_version_090(self, app):
        resp = app.get(
            "http://localhost/service?SERVICE=WMS&REQUEST=GetCapabilities"
            "&WMTVER=0.9.0"
        )
        assert is_110_capa(resp.lxml)

    def test_unsupported_version_130(self, app):
        resp = app.get(
            "http://localhost/service?SERVICE=WMS&REQUEST=GetCapabilities"
            "&VERSION=1.3.0"
        )
        assert is_111_capa(resp.lxml)

    def test_unknown_version_200(self, app):
        resp = app.get(
            "http://localhost/service?SERVICE=WMS&REQUEST=GetCapabilities"
            "&VERSION=2.0.0"
        )
        assert is_111_capa(resp.lxml)


class TestWMSVersionsTest(SysTest):

    @pytest.fixture(scope="class")
    def config_file(self):
        return "layer.yaml"

    def test_default_version_130(self, app):
        resp = app.get("http://localhost/service?SERVICE=WMS&REQUEST=GetCapabilities")
        assert is_130_capa(resp.lxml)

    def test_unknown_version_200(self, app):
        resp = app.get(
            "http://localhost/service?SERVICE=WMS&REQUEST=GetCapabilities"
            "&VERSION=2.0.0"
        )
        assert is_130_capa(resp.lxml)



# This file is part of the MapProxy project.
# Copyright (C) 2011 Omniscale <http://omniscale.de>
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import division

import functools

from io import BytesIO

import pytest

from mapproxy.test.image import is_jpeg, create_tmp_image
from mapproxy.test.http import MockServ
from mapproxy.test.helper import validate_with_xsd
from mapproxy.test.system import SysTest


@pytest.fixture(scope="module")
def config_file():
    return "wmts.yaml"


ns_wmts = {
    "wmts": "http://www.opengis.net/wmts/1.0",
    "ows": "http://www.opengis.net/ows/1.1",
    "xlink": "http://www.w3.org/1999/xlink",
}


def assert_xpath(xml, xpath, expected, namespaces=None):
    assert xml.xpath(xpath, namespaces=namespaces)[0] == expected


assert_xpath_wmts = functools.partial(assert_xpath, namespaces=ns_wmts)


class TestWMTS(SysTest):

    def test_capabilities(self, app):
        resp = app.get("/wmts/1.0.0/WMTSCapabilities.xml")
        xml = resp.lxml
        assert validate_with_xsd(
            xml, xsd_name="wmts/1.0/wmtsGetCapabilities_response.xsd"
        )
        assert len(xml.xpath("//wmts:Layer", namespaces=ns_wmts)) == 5
        assert (
            len(xml.xpath("//wmts:Contents/wmts:TileMatrixSet", namespaces=ns_wmts))
            == 5
        )
        # check InfoFormat for queryable layers
        for layer in xml.xpath("//wmts:Layer", namespaces=ns_wmts):
            if layer.findtext("ows:Identifier", namespaces=ns_wmts) in (
                "wms_cache",
                "wms_cache_multi",
                "gk3_cache",
                "tms_cache",
            ):
                assert layer.xpath("wmts:InfoFormat/text()", namespaces=ns_wmts) == [
                    "application/gml+xml; version=3.1",
                    "application/json",
                ]
            else:
                assert layer.xpath("wmts:InfoFormat/text()", namespaces=ns_wmts) == []

        # check ResourceURL for wms_cache
        layer = xml.xpath(
            '//wmts:Layer[ows:Identifier/text()="wms_cache"]', namespaces=ns_wmts
        )[0]
        resourceURLs = layer.xpath("wmts:ResourceURL", namespaces=ns_wmts)

        for rurl, format, type, template in [
            [
                resourceURLs[0],
                "image/jpeg",
                "tile",
                "http://localhost/wmts/myrest/wms_cache/{TileMatrixSet}/{TileMatrix}/{TileCol}/{TileRow}.jpeg",
            ],
            [
                resourceURLs[1],
                "application/gml+xml; version=3.1",
                "FeatureInfo",
                "http://localhost/wmts/myrest/wms_cache/{TileMatrixSet}/{TileMatrix}/{TileCol}/{TileRow}/{I}/{J}.gml",
            ],
            [
                resourceURLs[2],
                "application/json",
                "FeatureInfo",
                "http://localhost/wmts/myrest/wms_cache/{TileMatrixSet}/{TileMatrix}/{TileCol}/{TileRow}/{I}/{J}.geojson",  # noqa
            ],
        ]:
            assert rurl.attrib["format"] == format
            assert rurl.attrib["resourceType"] == type
            assert rurl.attrib["template"] == template

    def test_get_tile(self, app, fixture_cache_data):
        resp = app.get("/wmts/myrest/wms_cache/GLOBAL_MERCATOR/01/0/0.jpeg")
        assert resp.content_type == "image/jpeg"
        data = BytesIO(resp.body)
        assert is_jpeg(data)
        # test without leading 0 in level
        resp = app.get("/wmts/myrest/wms_cache/GLOBAL_MERCATOR/1/0/0.jpeg")
        assert resp.content_type == "image/jpeg"
        data = BytesIO(resp.body)
        assert is_jpeg(data)

    def test_get_tile_flipped_axis(self, app):
        serv = MockServ(port=42423)
        # source is ll, cache/service ul
        serv.expects("/tiles/01/000/000/000/000/000/001.png")
        serv.returns(create_tmp_image((256, 256)))
        with serv:
            resp = app.get("/wmts/myrest/tms_cache_ul/ulgrid/01/0/0.png", status=200)
            assert resp.content_type == "image/png"
            # test without leading 0 in level
            resp = app.get("/wmts/myrest/tms_cache_ul/ulgrid/1/0/0.png", status=200)
            assert resp.content_type == "image/png"

    def test_get_tile_source_error(self, app):
        resp = app.get("/wmts/myrest/tms_cache/GLOBAL_MERCATOR/01/0/0.png", status=500)
        xml = resp.lxml
        assert validate_with_xsd(xml, xsd_name="ows/1.1.0/owsExceptionReport.xsd")
        assert_xpath_wmts(
            xml, "/ows:ExceptionReport/ows:Exception/@exceptionCode", "NoApplicableCode"
        )

    def test_get_tile_out_of_range(self, app):
        resp = app.get(
            "/wmts/myrest/wms_cache/GLOBAL_MERCATOR/01/-1/0.jpeg", status=400
        )
        xml = resp.lxml
        assert resp.content_type == "text/xml"
        assert validate_with_xsd(xml, xsd_name="ows/1.1.0/owsExceptionReport.xsd")
        assert_xpath_wmts(
            xml, "/ows:ExceptionReport/ows:Exception/@exceptionCode", "TileOutOfRange"
        )

    def test_get_tile_invalid_format(self, app):
        url = "/wmts/myrest/wms_cache/GLOBAL_MERCATOR/01/0/0.png"
        self.check_invalid_parameter(app, url)

    def test_get_tile_invalid_layer(self, app):
        url = "/wmts/myrest/unknown/GLOBAL_MERCATOR/01/0/0.jpeg"
        self.check_invalid_parameter(app, url)

    def test_get_tile_invalid_matrixset(self, app):
        url = "/wmts/myrest/wms_cache/unknown/01/0/0.jpeg"
        self.check_invalid_parameter(app, url)

    def check_invalid_parameter(self, app, url):
        resp = app.get(url, status=400)
        xml = resp.lxml
        assert resp.content_type == "text/xml"
        assert validate_with_xsd(xml, xsd_name="ows/1.1.0/owsExceptionReport.xsd")
        assert_xpath_wmts(
            xml,
            "/ows:ExceptionReport/ows:Exception/@exceptionCode",
            "InvalidParameterValue",
        )

    def test_getfeatureinfo(self, app):
        fi_req = '/wmts/myrest/wms_cache/GLOBAL_MERCATOR/00/0/0/17/27.geojson'
        serv = MockServ(port=42423)
        serv.expects(
            "/service?layers=foo,bar"
            + "&bbox=-20037508.342789244,-20037508.342789244,20037508.342789244,20037508.342789244"
            + "&width=256&height=256&x=17&y=27&query_layers=foo,bar&format=image%2Fpng&srs=EPSG%3A900913"
            + "&request=GetFeatureInfo&version=1.1.1&service=WMS&styles=&info_format=application/json"
        )
        serv.returns(b'{"data": 43}')
        with serv:
            resp = app.get(fi_req, status=200)
            assert resp.content_type == "application/json"

    def test_getfeatureinfo_xml(self, app):
        fi_req = '/wmts/myrest/wms_cache/GLOBAL_MERCATOR/00/0/0/17/27.gml'
        serv = MockServ(port=42423)
        serv.expects(
            "/service?layers=foo,bar"
            + "&bbox=-20037508.342789244,-20037508.342789244,20037508.342789244,20037508.342789244"
            + "&width=256&height=256&x=17&y=27&query_layers=foo,bar&format=image%2Fpng&srs=EPSG%3A900913"
            + "&request=GetFeatureInfo&version=1.1.1&service=WMS&styles=&info_format=application/gml%2bxml%3b%20version=3.1"  # noqa
        )
        serv.returns(b"<root />")
        with serv:
            resp = app.get(str(fi_req), status=200)
            assert resp.headers["Content-type"] == "application/gml+xml; version=3.1"



# This file is part of the MapProxy project.
# Copyright (C) 2010 Omniscale <http://omniscale.de>
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import print_function

import os
from itertools import zip_longest

from mapproxy.seed.util import format_cleanup_task
from mapproxy.util.fs import cleanup_directory
from mapproxy.seed.seeder import (
    TileWorkerPool, TileWalker, TileCleanupWorker,
    SeedProgress,
)


def cleanup(tasks, concurrency=2, dry_run=False, skip_geoms_for_last_levels=0,
            verbose=True, progress_logger=None):
    for task in tasks:
        print(format_cleanup_task(task))

        if task.coverage is False:
            continue

        # seed_progress for tilewalker cleanup
        seed_progress = None
        # cleanup_progress for os.walk based cleanup
        cleanup_progress = None
        if progress_logger and progress_logger.progress_store:
            progress_logger.current_task_id = task.id
            start_progress = progress_logger.progress_store.get(task.id)
            seed_progress = SeedProgress(old_progress_identifier=start_progress)
            cleanup_progress = DirectoryCleanupProgress(old_dir=start_progress)

        if task.complete_extent:
            if callable(getattr(task.tile_manager.cache, 'level_location', None)):
                simple_cleanup(task, dry_run=dry_run, progress_logger=progress_logger,
                               cleanup_progress=cleanup_progress)
                task.tile_manager.cleanup()
                continue
            elif callable(getattr(task.tile_manager.cache, 'remove_level_tiles_before', None)):
                cache_cleanup(task, dry_run=dry_run, progress_logger=progress_logger)
                task.tile_manager.cleanup()
                continue

        tilewalker_cleanup(task, dry_run=dry_run, concurrency=concurrency,
                           skip_geoms_for_last_levels=skip_geoms_for_last_levels,
                           progress_logger=progress_logger,
                           seed_progress=seed_progress,
                           )
        task.tile_manager.cleanup()


def simple_cleanup(task, dry_run, progress_logger=None, cleanup_progress=None):
    """
    Cleanup cache level on file system level.
    """

    for level in task.levels:
        level_dir = task.tile_manager.cache.level_location(level)
        if dry_run:
            def file_handler(filename):
                print('removing ' + filename)
        else:
            file_handler = None
        if progress_logger:
            progress_logger.log_message('removing old tiles in ' + normpath(level_dir))
            if progress_logger.progress_store:
                cleanup_progress.step_dir(level_dir)
                if cleanup_progress.already_processed():
                    continue
                progress_logger.progress_store.add(
                    task.id,
                    cleanup_progress.current_progress_identifier(),
                )
                progress_logger.progress_store.write()

        cleanup_directory(level_dir, task.remove_timestamp, task.remove_all,
                          file_handler=file_handler, remove_empty_dirs=True)


def cache_cleanup(task, dry_run, progress_logger=None):
    for level in task.levels:
        if progress_logger:
            progress_logger.log_message('removing old tiles for level %s' % level)
        if not dry_run:
            task.tile_manager.cache.remove_level_tiles_before(level, task.remove_timestamp, task.remove_all)
            task.tile_manager.cleanup()


def normpath(path):
    # relpath doesn't support UNC
    if path.startswith('\\'):
        return path

    path = os.path.relpath(path)

    if path.startswith('../../'):
        path = os.path.abspath(path)
    return path


def tilewalker_cleanup(task, dry_run, concurrency, skip_geoms_for_last_levels,
                       progress_logger=None, seed_progress=None):
    """
    Cleanup tiles with tile traversal.
    """
    if task.remove_all:
        handle_all = True
    else:
        task.tile_manager._expire_timestamp = task.remove_timestamp
        handle_all = False

    task.tile_manager.minimize_meta_requests = False
    tile_worker_pool = TileWorkerPool(task, TileCleanupWorker, progress_logger=progress_logger,
                                      dry_run=dry_run, size=concurrency)
    tile_walker = TileWalker(task, tile_worker_pool, handle_stale=True, handle_all=handle_all,
                             work_on_metatiles=False, progress_logger=progress_logger,
                             skip_geoms_for_last_levels=skip_geoms_for_last_levels,
                             seed_progress=seed_progress)
    try:
        tile_walker.walk()
    except KeyboardInterrupt:
        tile_worker_pool.stop(force=True)
        raise
    finally:
        tile_worker_pool.stop()


class DirectoryCleanupProgress(object):
    def __init__(self, old_dir=None):
        self.old_dir = old_dir
        self.current_dir = None

    def step_dir(self, dir):
        self.current_dir = dir

    def already_processed(self):
        return self.can_skip(self.old_dir, self.current_dir)

    def current_progress_identifier(self):
        if self.already_processed() or self.current_dir is None:
            return self.old_dir
        return self.current_dir

    @staticmethod
    def can_skip(old_dir, current_dir):
        """
        Return True if the `current_dir` is before `old_dir` when compared
        lexicographic.

        >>> DirectoryCleanupProgress.can_skip(None, '/00')
        False
        >>> DirectoryCleanupProgress.can_skip(None, '/00/000/000')
        False

        >>> DirectoryCleanupProgress.can_skip('/01/000/001', '/00')
        True
        >>> DirectoryCleanupProgress.can_skip('/01/000/001', '/01/000/000')
        True
        >>> DirectoryCleanupProgress.can_skip('/01/000/001', '/01/000/000/000')
        True
        >>> DirectoryCleanupProgress.can_skip('/01/000/001', '/01/000/001')
        False
        >>> DirectoryCleanupProgress.can_skip('/01/000/001', '/01/000/001/000')
        False
        """
        if old_dir is None:
            return False
        if current_dir is None:
            return False
        for old, current in zip_longest(old_dir.split(os.path.sep), current_dir.split(os.path.sep), fillvalue=None):
            if old is None:
                return False
            if current is None:
                return False
            if old < current:
                return False
            if old > current:
                return True
        return False

    def running(self):
        return True



# -:- encoding: utf-8 -:-
# This file is part of the MapProxy project.
# Copyright (C) 2013 Omniscale <http://omniscale.de>
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


def seeds(cap, caches):
    seeds = {}
    cleanups = {}

    for cache_name, cache in caches.items():
        for grid in cache['grids']:
            seeds[cache_name + '_' + grid] = {
                'caches': [cache_name],
                'grids': [grid],
            }
            cleanups[cache_name + '_' + grid] = {
                'caches': [cache_name],
                'grids': [grid],
                'remove_before': {
                    'time': '1900-01-01T00:00:00',
                }
            }

    return seeds, cleanups



# This file is part of the MapProxy project.
# Copyright (C) 2010-2012 Omniscale <http://omniscale.de>
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import os
import hashlib

from io import BytesIO

import pytest

from PIL import Image
from mapproxy.test.image import is_jpeg, tmp_image
from mapproxy.test.http import mock_httpd
from mapproxy.test.system import SysTest


@pytest.fixture(scope="module")
def config_file():
    return "layer.yaml"


class TestTMS(SysTest):

    def test_tms_capabilities(self, app):
        resp = app.get("/tms/1.0.0/")
        assert "WMS Cache Layer" in resp
        assert "WMS Cache Multi Layer" in resp
        assert "TMS Cache Layer" in resp
        assert "TMS Cache Layer + FI" in resp
        xml = resp.lxml
        assert xml.xpath("count(//TileMap)") == 11
        assert "Bob Mustermann" in resp
        assert "keyword1" in resp

        # without trailing space
        resp2 = app.get("/tms/1.0.0")
        assert resp.body == resp2.body

    def test_tms_layer_capabilities(self, app):
        resp = app.get("/tms/1.0.0/wms_cache")
        assert "WMS Cache Layer" in resp
        xml = resp.lxml
        assert xml.xpath("count(//TileSet)") == 19

    def test_tms_root_resource(self, app):
        resp = app.get("/tms")
        resp2 = app.get("/tms/")
        assert "TileMapService" in resp and "TileMapService" in resp2
        xml = resp.lxml
        assert xml.xpath("//TileMapService/@version") == ["1.0.0"]

    def test_tms_get_out_of_bounds_tile(self, app):
        for coord in [(0, 0, -1), (-1, 0, 0), (0, -1, 0), (4, 2, 1), (1, 3, 0)]:
            x, y, z = coord
            url = "/tms/1.0.0/wms_cache/%d/%d/%d.jpeg" % (z, x, y)
            resp = app.get(url, status=404)
            xml = resp.lxml
            assert (
                "outside the bounding box"
                in xml.xpath("/TileMapServerError/Message/text()")[0]
            )

    def test_invalid_layer(self, app):
        resp = app.get("/tms/1.0.0/inVAlid/0/0/0.png", status=404)
        xml = resp.lxml
        assert (
            "unknown layer: inVAlid"
            in xml.xpath("/TileMapServerError/Message/text()")[0]
        )

    def test_invalid_format(self, app):
        resp = app.get("/tms/1.0.0/wms_cache/0/0/1.png", status=404)
        xml = resp.lxml
        assert "invalid format" in xml.xpath("/TileMapServerError/Message/text()")[0]

    def test_get_tile_tile_source_error(self, app):
        resp = app.get("/tms/1.0.0/wms_cache/0/0/0.jpeg", status=500)
        xml = resp.lxml
        assert (
            "No response from URL" in xml.xpath("/TileMapServerError/Message/text()")[0]
        )

    def test_get_cached_tile(self, app, fixture_cache_data):
        resp = app.get("/tms/1.0.0/wms_cache/0/0/1.jpeg")
        assert resp.content_type == "image/jpeg"
        assert resp.content_length == len(resp.body)
        data = BytesIO(resp.body)
        assert is_jpeg(data)

    def test_get_tile(self, app, cache_dir):
        with tmp_image((256, 256), format="jpeg") as img:
            expected_req = (
                {
                    "path": r"/service?LAYERs=foo,bar&SERVICE=WMS&FORMAT=image%2Fjpeg"
                    "&REQUEST=GetMap&HEIGHT=256&SRS=EPSG%3A900913&styles="
                    "&VERSION=1.1.1&BBOX=-20037508.3428,-20037508.3428,0.0,0.0"
                    "&WIDTH=256"
                },
                {"body": img.read(), "headers": {"content-type": "image/jpeg"}},
            )
            with mock_httpd(
                ("localhost", 42423), [expected_req], bbox_aware_query_comparator=True
            ):
                resp = app.get("/tms/1.0.0/wms_cache/0/0/0.jpeg")
                assert resp.content_type == "image/jpeg"
        assert cache_dir.join(
            "wms_cache_EPSG900913/01/000/000/000/000/000/000.jpeg"
        ).check()

    def test_get_tile_from_cache_with_tile_source(self, app, cache_dir):
        with tmp_image((256, 256), format="jpeg") as img:
            expected_req = (
                {"path": r"/tiles/01/000/000/000/000/000/001.png"},
                {"body": img.read(), "headers": {"content-type": "image/png"}},
            )
            with mock_httpd(("localhost", 42423), [expected_req]):
                resp = app.get("/tms/1.0.0/tms_cache/0/0/1.png")
                assert resp.content_type == "image/png"
        assert cache_dir.join(
            "tms_cache_EPSG900913/01/000/000/000/000/000/001.png"
        ).check()

    def test_get_tile_with_watermark_cache(self, app):
        with tmp_image((256, 256), format="png", color=(0, 0, 0)) as img:
            expected_req = (
                {"path": r"/tiles/01/000/000/000/000/000/000.png"},
                {"body": img.read(), "headers": {"content-type": "image/png"}},
            )
            with mock_httpd(("localhost", 42423), [expected_req]):
                resp = app.get("/tms/1.0.0/watermark_cache/0/0/0.png")
                assert resp.content_type == "image/png"
                img = Image.open(BytesIO(resp.body))
                colors = img.getcolors()
                assert len(colors) >= 2
                assert sorted(colors)[-1][1] == (0, 0, 0)


class TestTileService(SysTest):

    def test_get_out_of_bounds_tile(self, app):
        for coord in [(0, 0, -1), (-1, 0, 0), (0, -1, 0), (4, 2, 1), (1, 3, 0)]:
            x, y, z = coord
            url = "/tiles/wms_cache/%d/%d/%d.jpeg" % (z, x, y)
            resp = app.get(url, status=404)
            assert "outside the bounding box" in resp

    def test_invalid_layer(self, app):
        resp = app.get("/tiles/inVAlid/0/0/0.png", status=404)
        assert resp.content_type == "text/plain"
        assert "unknown layer: inVAlid" in resp

    def test_invalid_format(self, app):
        resp = app.get("/tiles/wms_cache/0/0/1.png", status=404)
        assert resp.content_type == "text/plain"
        assert "invalid format" in resp

    def test_get_tile_tile_source_error(self, app):
        resp = app.get("/tiles/wms_cache/0/0/0.jpeg", status=500)
        assert resp.content_type == "text/plain"
        assert "No response from URL" in resp

    def _check_tile_resp(self, resp):
        assert resp.content_type == "image/jpeg"
        assert resp.content_length == len(resp.body)
        data = BytesIO(resp.body)
        assert is_jpeg(data)

    def _update_timestamp(self, cache_dir, base_config):
        timestamp = 1234567890.0
        size = 10214
        base_dir = base_config.cache.base_dir
        os.utime(
            os.path.join(
                base_dir, "wms_cache_EPSG900913/01/000/000/000/000/000/001.jpeg"
            ),
            (timestamp, timestamp),
        )
        max_age = base_config.tiles.expires_hours * 60 * 60
        md5 = hashlib.new('md5', (str(timestamp) + str(size)).encode("ascii"), usedforsecurity=False)
        etag = md5.hexdigest()
        return etag, max_age

    def _check_cache_control_headers(self, resp, etag, max_age):
        assert resp.headers["ETag"] == etag
        assert resp.headers["Last-modified"] == "Fri, 13 Feb 2009 23:31:30 GMT"
        assert resp.headers["Cache-control"] == "public, max-age=%d, s-maxage=%d" % (
            max_age,
            max_age,
        )

    def test_get_cached_tile(self, app, cache_dir, base_config, fixture_cache_data):
        etag, max_age = self._update_timestamp(cache_dir, base_config)
        resp = app.get("/tiles/wms_cache/1/0/1.jpeg")
        self._check_cache_control_headers(resp, etag, max_age)
        self._check_tile_resp(resp)

    def test_get_cached_tile_flipped_y(
        self, app, cache_dir, base_config, fixture_cache_data
    ):
        etag, max_age = self._update_timestamp(cache_dir, base_config)
        resp = app.get("/tiles/wms_cache/1/0/0.jpeg?origin=nw")
        self._check_cache_control_headers(resp, etag, max_age)
        self._check_tile_resp(resp)

    def test_if_none_match(self, app, cache_dir, base_config, fixture_cache_data):
        etag, max_age = self._update_timestamp(cache_dir, base_config)
        resp = app.get("/tiles/wms_cache/1/0/1.jpeg", headers={"If-None-Match": etag})
        assert resp.status == "304 Not Modified"
        self._check_cache_control_headers(resp, etag, max_age)

        resp = app.get(
            "/tiles/wms_cache/1/0/1.jpeg", headers={"If-None-Match": etag + "foo"}
        )
        self._check_cache_control_headers(resp, etag, max_age)
        assert resp.status == "200 OK"
        self._check_tile_resp(resp)

    @pytest.mark.parametrize(
        "date,modified",
        [
            ("Fri, 15 Feb 2009 23:31:30 GMT", False),
            ("Fri, 13 Feb 2009 23:31:31 GMT", False),
            ("Fri, 13 Feb 2009 23:31:30 GMT", False),
            ("Fri, 13 Feb 2009 23:31:29 GMT", True),
            ("Fri, 11 Feb 2009 23:31:29 GMT", True),
            ("Friday, 13-Feb-09 23:31:30 GMT", False),
            ("Friday, 13-Feb-09 23:31:29 GMT", True),
            ("Fri Feb 13 23:31:30 2009", False),
            ("Fri Feb 13 23:31:29 2009", True),
            # and some invalid ones
            ("Fri Foo 13 23:31:29 2009", True),
            ("1234567890", True),
        ],
    )
    def test_if_modified_since(
        self, app, cache_dir, base_config, fixture_cache_data, date, modified
    ):
        etag, max_age = self._update_timestamp(cache_dir, base_config)
        resp = app.get(
            "/tiles/wms_cache/1/0/1.jpeg", headers={"If-Modified-Since": date}
        )
        self._check_cache_control_headers(resp, etag, max_age)
        if modified:
            assert resp.status == "200 OK"
            self._check_tile_resp(resp)
        else:
            assert resp.status == "304 Not Modified"

    def test_get_tile(self, app, cache_dir):
        with tmp_image((256, 256), format="jpeg") as img:
            expected_req = (
                {
                    "path": r"/service?LAYERs=foo,bar&SERVICE=WMS&FORMAT=image%2Fjpeg"
                    "&REQUEST=GetMap&HEIGHT=256&SRS=EPSG%3A900913&styles="
                    "&VERSION=1.1.1&BBOX=-20037508.3428,-20037508.3428,0.0,0.0"
                    "&WIDTH=256"
                },
                {"body": img.read(), "headers": {"content-type": "image/jpeg"}},
            )
            with mock_httpd(
                ("localhost", 42423), [expected_req], bbox_aware_query_comparator=True
            ):
                resp = app.get("/tiles/wms_cache/1/0/0.jpeg")
                assert resp.content_type == "image/jpeg"
        assert cache_dir.join(
            "wms_cache_EPSG900913/01/000/000/000/000/000/000.jpeg"
        ).check()



# This file is part of the MapProxy project.
# Copyright (C) 2010, 2011 Omniscale <http://omniscale.de>
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import print_function, division

import os
import sys
import stat
import math
import time
from datetime import datetime

try:
    import cPickle as pickle
except ImportError:
    import pickle

from mapproxy.layer import map_extent_from_grid
from mapproxy.util.fs import write_atomic

import logging
log = logging.getLogger(__name__)


class bidict(dict):
    """
    Simplest bi-directional dictionary.
    """

    def __init__(self, iterator):
        for key, val in iterator:
            dict.__setitem__(self, key, val)
            dict.__setitem__(self, val, key)


class ProgressStore(object):
    """
    Reads and stores seed progresses to a file.
    """

    def __init__(self, filename=None, continue_seed=True):
        self.filename = filename
        if continue_seed:
            self.status = self.load()
        else:
            self.status = {}

    def load(self):
        if not os.path.exists(self.filename):
            pass
        elif os.stat(self.filename).st_mode & stat.S_IWOTH:
            log.error('progress file (%s) is world writable, ignoring file',
                      self.filename)
        else:
            with open(self.filename, 'rb') as f:
                try:
                    return pickle.load(f)
                except (pickle.UnpicklingError, AttributeError,
                        EOFError, ImportError, IndexError):
                    log.error('unable to read progress file (%s), ignoring file',
                              self.filename)

        return {}

    def write(self):
        try:
            write_atomic(self.filename, pickle.dumps(self.status))
        except (IOError, OSError) as ex:
            log.error('unable to write seed progress: %s', ex)

    def remove(self):
        self.status = {}
        if os.path.exists(self.filename):
            os.remove(self.filename)

    def get(self, task_identifier):
        return self.status.get(task_identifier, None)

    def add(self, task_identifier, progress_identifier):
        self.status[task_identifier] = progress_identifier


class ProgressLog(object):
    def __init__(self, out=None, silent=False, verbose=True, progress_store=None):
        if not out:
            out = sys.stdout
        self.out = out
        self._laststep = time.time()
        self._lastprogress = 0

        self.verbose = verbose
        self.silent = silent
        self.current_task_id = None
        self.progress_store = progress_store

    def log_message(self, msg):
        self.out.write('[%s] %s\n' % (
            timestamp(), msg,
        ))
        self.out.flush()

    def log_step(self, progress):
        if not self.verbose:
            return
        if (self._laststep + .5) < time.time():
            # log progress at most every 500ms
            self.out.write('[%s] %6.2f%%\t%-20s \r' % (
                timestamp(), progress.progress*100, progress.progress_str,
            ))
            self.out.flush()
            self._laststep = time.time()

    def log_progress(self, progress, level, bbox, tiles):
        progress_interval = 1
        if not self.verbose:
            progress_interval = 30

        log_progess = False
        if progress.progress == 1.0 or (self._lastprogress + progress_interval) < time.time():
            self._lastprogress = time.time()
            log_progess = True

        if log_progess:
            if self.progress_store and self.current_task_id:
                self.progress_store.add(self.current_task_id,
                                        progress.current_progress_identifier())
                self.progress_store.write()

        if self.silent:
            return

        if log_progess:
            self.out.write('[%s] %2s %6.2f%% %s (%d tiles)\n' % (
                timestamp(), level, progress.progress*100,
                format_bbox(bbox), tiles))
            self.out.flush()


def limit_sub_bbox(bbox, sub_bbox):
    """
    >>> limit_sub_bbox((0, 1, 10, 11), (-1, -1, 9, 8))
    (0, 1, 9, 8)
    >>> limit_sub_bbox((0, 0, 10, 10), (5, 2, 18, 18))
    (5, 2, 10, 10)
    """
    minx = max(bbox[0], sub_bbox[0])
    miny = max(bbox[1], sub_bbox[1])
    maxx = min(bbox[2], sub_bbox[2])
    maxy = min(bbox[3], sub_bbox[3])
    return minx, miny, maxx, maxy


def timestamp():
    return datetime.now().strftime('%H:%M:%S')


def format_bbox(bbox):
    return ('%.5f, %.5f, %.5f, %.5f') % tuple(bbox)


def status_symbol(i, total):
    """
    >>> status_symbol(0, 1)
    '0'
    >>> [status_symbol(i, 4) for i in range(5)]
    ['.', 'o', 'O', '0', 'X']
    >>> [status_symbol(i, 10) for i in range(11)]
    ['.', '.', 'o', 'o', 'o', 'O', 'O', '0', '0', '0', 'X']
    """
    symbols = list(' .oO0')
    i += 1
    if 0 < i > total:
        return 'X'
    else:
        return symbols[int(math.ceil(i/(total/4)))]


class BackoffError(Exception):
    pass


def exp_backoff(func, args=(), kw={}, max_repeat=10, start_backoff_sec=2,
                exceptions=(Exception,), ignore_exceptions=tuple(), max_backoff=60):
    n = 0
    while True:
        try:
            result = func(*args, **kw)
        except ignore_exceptions:
            time.sleep(0.01)
        except exceptions as ex:
            if n >= max_repeat:
                print("An error occured. Giving up", file=sys.stderr)
                raise BackoffError
            wait_for = start_backoff_sec * 2**n
            if wait_for > max_backoff:
                wait_for = max_backoff
            print("An error occured. Retry in %d seconds: %r. Retries left: %d" %
                  (wait_for, ex, (max_repeat - n)), file=sys.stderr)
            time.sleep(wait_for)
            n += 1
        else:
            return result


def format_seed_task(task):
    info = []
    info.append('  %s:' % (task.md['name'], ))
    if task.coverage is False:
        info.append("    Empty coverage given for this task")
        info.append("    Skipped")
        return '\n'.join(info)

    info.append("    Seeding cache '%s' with grid '%s' in %s" % (
        task.md['cache_name'], task.md['grid_name'], task.grid.srs.srs_code))
    if task.coverage:
        info.append('    Limited to coverage in: %s (EPSG:4326)' % (format_bbox(task.coverage.extent.llbbox), ))
    else:
        info.append('   Complete grid: %s (EPSG:4326)' % (format_bbox(map_extent_from_grid(task.grid).llbbox), ))
    info.append('    Levels: %s' % (task.levels, ))

    if task.refresh_all:
        info.append('    Overwriting: all tiles')
    elif task.refresh_timestamp:
        info.append('    Overwriting: tiles older than %s' %
                    datetime.fromtimestamp(task.refresh_timestamp))
    else:
        info.append('    Overwriting: no tiles')

    return '\n'.join(info)


def format_cleanup_task(task):
    info = []
    info.append('  %s:' % (task.md['name'], ))
    if task.coverage is False:
        info.append("    Empty coverage given for this task")
        info.append("    Skipped")
        return '\n'.join(info)

    info.append("    Cleaning up cache '%s' with grid '%s' in %s" % (
        task.md['cache_name'], task.md['grid_name'], task.grid.srs.srs_code))
    if task.coverage:
        info.append('    Limited to coverage in: %s (EPSG:4326)' % (format_bbox(task.coverage.extent.llbbox), ))
    else:
        info.append('    Complete grid: %s (EPSG:4326)' % (format_bbox(map_extent_from_grid(task.grid).llbbox), ))
    info.append('    Levels: %s' % (task.levels, ))

    if task.remove_all:
        info.append('    Remove: all tiles')
    else:
        info.append('    Remove: tiles older than %s' %
                    datetime.fromtimestamp(task.remove_timestamp))

    return '\n'.join(info)



# This file is part of the MapProxy project.
# Copyright (C) 2010-2014 Omniscale <http://omniscale.de>
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import division

import pytest

from mapproxy.request.wms import WMS111MapRequest
from mapproxy.test.image import (
    is_transparent,
    create_tmp_image,
    bgcolor_ratio,
    img_from_buf,
    assert_colors_equal,
)
from mapproxy.test.http import (
    assert_no_cache,
    mock_httpd
)
from mapproxy.test.system import SysTest
from mapproxy.test.system.test_wms import is_111_exception


transp = create_tmp_image((200, 200), mode="RGBA", color=(0, 0, 0, 0))


class TestWMS(SysTest):

    @pytest.fixture(scope="class")
    def config_file(self):
        return "source_errors.yaml"

    def setup_method(self):
        self.common_map_req = WMS111MapRequest(
            url="/service?",
            param=dict(
                service="WMS",
                version="1.1.1",
                bbox="9,50,10,51",
                width="200",
                height="200",
                layers="online",
                srs="EPSG:4326",
                format="image/png",
                styles="",
                request="GetMap",
                transparent=True,
            ),
        )

    def test_online(self, app):
        common_params = (
            r"?SERVICE=WMS&FORMAT=image%2Fpng"
            "&REQUEST=GetMap&HEIGHT=200&SRS=EPSG%3A4326&styles="
            "&VERSION=1.1.1&BBOX=9.0,50.0,10.0,51.0"
            "&WIDTH=200&transparent=True"
        )

        expected_req = [
            (
                {"path": "/service_a" + common_params + "&layers=a_one"},
                {"body": transp, "headers": {"content-type": "image/png"}},
            )
        ]

        with mock_httpd(("localhost", 42423), expected_req):
            self.common_map_req.params.layers = "online"
            resp = app.get(self.common_map_req)
            assert "Cache-Control" not in resp.headers
            assert resp.content_type == "image/png"
            assert is_transparent(resp.body)

    def test_mixed_layer_source(self, app):
        common_params = (
            r"?SERVICE=WMS&FORMAT=image%2Fpng"
            "&REQUEST=GetMap&HEIGHT=200&SRS=EPSG%3A4326&styles="
            "&VERSION=1.1.1&BBOX=9.0,50.0,10.0,51.0"
            "&WIDTH=200&transparent=True"
        )

        expected_req = [
            (
                {"path": "/service_a" + common_params + "&layers=a_one"},
                {"body": transp, "headers": {"content-type": "image/png"}},
            )
        ]

        with mock_httpd(("localhost", 42423), expected_req):
            self.common_map_req.params.layers = "mixed"
            resp = app.get(self.common_map_req)
            assert_no_cache(resp)
            assert resp.content_type == "image/png"
            assert 0.99 > bgcolor_ratio(resp.body) > 0.95

    def test_mixed_sources(self, app):
        common_params = (
            r"?SERVICE=WMS&FORMAT=image%2Fpng"
            "&REQUEST=GetMap&HEIGHT=200&SRS=EPSG%3A4326&styles="
            "&VERSION=1.1.1&BBOX=9.0,50.0,10.0,51.0"
            "&WIDTH=200&transparent=True"
        )

        expected_req = [
            (
                {"path": "/service_a" + common_params + "&layers=a_one"},
                {"body": transp, "headers": {"content-type": "image/png"}},
            )
        ]

        with mock_httpd(("localhost", 42423), expected_req):
            self.common_map_req.params.layers = "online,all_offline"
            resp = app.get(self.common_map_req)
            assert_no_cache(resp)
            assert resp.content_type == "image/png"
            assert 0.99 > bgcolor_ratio(resp.body) > 0.95
            # open('/tmp/foo.png', 'wb').write(resp.body)

    def test_all_offline(self, app):
        self.common_map_req.params.layers = "all_offline"
        resp = app.get(self.common_map_req, expect_errors=True)
        assert resp.content_type == "application/vnd.ogc.se_xml"
        is_111_exception(resp.lxml, re_msg="no response from url")


class TestWMSRaise(SysTest):

    @pytest.fixture(scope="class")
    def config_file(self):
        return "source_errors_raise.yaml"

    def setup_method(self):
        self.common_map_req = WMS111MapRequest(
            url="/service?",
            param=dict(
                service="WMS",
                version="1.1.1",
                bbox="9,50,10,51",
                width="200",
                height="200",
                layers="online",
                srs="EPSG:4326",
                format="image/png",
                styles="",
                request="GetMap",
                transparent=True,
            ),
        )

    def test_mixed_layer_source(self, app):
        common_params = (
            r"?SERVICE=WMS&FORMAT=image%2Fpng"
            "&REQUEST=GetMap&HEIGHT=200&SRS=EPSG%3A4326&styles="
            "&VERSION=1.1.1&BBOX=9.0,50.0,10.0,51.0"
            "&WIDTH=200&transparent=True"
        )

        expected_req = [
            (
                {"path": "/service_a" + common_params + "&layers=a_one"},
                {"body": transp, "headers": {"content-type": "image/png"}},
            )
        ]

        with mock_httpd(("localhost", 42423), expected_req):
            self.common_map_req.params.layers = "mixed"
            resp = app.get(self.common_map_req, expect_errors=True)
            is_111_exception(resp.lxml, re_msg="no response from url")

    def test_all_offline(self, app):
        self.common_map_req.params.layers = "all_offline"
        resp = app.get(self.common_map_req, expect_errors=True)
        assert resp.content_type == "application/vnd.ogc.se_xml"
        is_111_exception(resp.lxml, re_msg="no response from url")


class TestTileErrors(SysTest):

    @pytest.fixture(scope="class")
    def config_file(self):
        return "source_errors.yaml"

    def setup_method(self):
        self.common_map_req = WMS111MapRequest(
            url="/service?",
            param=dict(
                service="WMS",
                version="1.1.1",
                bbox="0,-90,180,90",
                width="250",
                height="250",
                layers="tilesource",
                srs="EPSG:4326",
                format="image/png",
                styles="",
                request="GetMap",
                transparent=True,
            ),
        )

        self.common_tile_req = "/tiles/tilesource/EPSG4326/1/1/0.png"

    def test_wms_uncached_response(self, app, cache_dir):
        expected_req = [
            (
                {"path": "/foo/1/1/0.png"},
                {
                    "body": b"not found",
                    "status": 404,
                    "headers": {"content-type": "text/plain"},
                },
            )
        ]

        with mock_httpd(("localhost", 42423), expected_req):
            resp = app.get(self.common_map_req)
            assert resp.content_type == "image/png"
            assert_no_cache(resp)
            img = img_from_buf(resp.body)
            assert img.getcolors() == [(250 * 250, (255, 0, 128))]
        assert not cache_dir.join(
            "tilesource_cache_EPSG4326/01/000/000/001/000/000/000.png"
        ).check()

    def test_wms_cached_response(self, app, cache_dir):
        expected_req = [
            (
                {"path": "/foo/1/1/0.png"},
                {
                    "body": b"no content",
                    "status": 204,
                    "headers": {"content-type": "text/plain"},
                },
            )
        ]

        with mock_httpd(("localhost", 42423), expected_req):
            resp = app.get(self.common_map_req)
            assert resp.content_type == "image/png"
            assert "Cache-Control" not in resp.headers
            img = img_from_buf(resp.body)
            assert_colors_equal(img, [(250 * 250, (100, 200, 50, 250))])
            assert cache_dir.join(
                "tilesource_cache_EPSG4326/01/000/000/001/000/000/000.png"
            ).check()

    def test_wms_unhandled_error_code(self, app):
        expected_req = [
            (
                {"path": "/foo/1/1/0.png"},
                {
                    "body": b"error",
                    "status": 500,
                    "headers": {"content-type": "text/plain"},
                },
            )
        ]

        with mock_httpd(("localhost", 42423), expected_req):
            resp = app.get(self.common_map_req, expect_errors=True)
            assert_no_cache(resp)
            assert resp.content_type == "application/vnd.ogc.se_xml"
            assert b"500" in resp.body

    def test_wms_catchall_error_no_image_response(self, app):
        expected_req = [
            (
                {"path": "/foo/1/1/0.png"},
                {
                    "body": b"error",
                    "status": 200,
                    "headers": {"content-type": "text/plain"},
                },
            )
        ]

        with mock_httpd(("localhost", 42423), expected_req):
            self.common_map_req.params["layers"] = "tilesource_catchall"
            resp = app.get(self.common_map_req)
            assert_no_cache(resp)
            assert resp.content_type == "image/png"
            img = img_from_buf(resp.body)
            assert img.getcolors() == [(250 * 250, (100, 50, 50))]

    def test_tile_uncached_response(self, app, cache_dir):
        expected_req = [
            (
                {"path": "/foo/1/1/0.png"},
                {
                    "body": b"not found",
                    "status": 404,
                    "headers": {"content-type": "text/plain"},
                },
            )
        ]

        with mock_httpd(("localhost", 42423), expected_req):
            resp = app.get(self.common_tile_req)
            assert_no_cache(resp)
            assert resp.content_type == "image/png"
            img = img_from_buf(resp.body)
            assert img.getcolors() == [(256 * 256, (255, 0, 128))]
            assert not cache_dir.join(
                "tilesource_cache_EPSG4326/01/000/000/001/000/000/000.png"
            ).check()

    def test_tile_cached_response(self, app, cache_dir):
        expected_req = [
            (
                {"path": "/foo/1/1/0.png"},
                {
                    "body": b"no content",
                    "status": 204,
                    "headers": {"content-type": "text/plain"},
                },
            )
        ]

        with mock_httpd(("localhost", 42423), expected_req):
            resp = app.get(self.common_tile_req)
            assert "public" in resp.headers["Cache-Control"]
            assert resp.content_type == "image/png"
            img = img_from_buf(resp.body)
            assert img.getcolors() == [(256 * 256, (100, 200, 50, 250))]
            assert cache_dir.join(
                "tilesource_cache_EPSG4326/01/000/000/001/000/000/000.png"
            ).check()

    def test_tile_unhandled_error_code(self, app):
        expected_req = [
            (
                {"path": "/foo/1/1/0.png"},
                {
                    "body": b"error",
                    "status": 500,
                    "headers": {"content-type": "text/plain"},
                },
            )
        ]

        with mock_httpd(("localhost", 42423), expected_req):
            resp = app.get(self.common_tile_req, status=500)
            assert_no_cache(resp)
            assert resp.content_type == "text/plain"
            assert b"500" in resp.body

    def test_tile_catchall_error_no_image_response(self, app):
        expected_req = [
            (
                {"path": "/foo/1/1/0.png"},
                {
                    "body": b"error",
                    "status": 200,
                    "headers": {"content-type": "text/plain"},
                },
            )
        ]

        with mock_httpd(("localhost", 42423), expected_req):
            resp = app.get(
                self.common_tile_req.replace("tilesource", "tilesource_catchall")
            )
            assert_no_cache(resp)
            assert resp.content_type == "image/png"
            img = img_from_buf(resp.body)
            assert img.getcolors() == [(256 * 256, (100, 50, 50))]



# This file is part of the MapProxy project.
# Copyright (C) 2010 Omniscale <http://omniscale.de>
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import pytest

from mapproxy.util.collections import LRU, ImmutableDictList


class TestLRU(object):
    def test_missing_key(self):
        lru = LRU(10)
        with pytest.raises(KeyError):
            lru['foo']

    def test_contains(self):
        lru = LRU(10)
        lru['foo1'] = 1

        assert 'foo1' in lru
        assert 'foo2' not in lru

    def test_repr(self):
        lru = LRU(10)
        lru['foo1'] = 1
        assert 'size=10' in repr(lru)
        assert 'foo1' in repr(lru)

    def test_getitem(self):
        lru = LRU(10)
        lru['foo1'] = 1
        lru['foo2'] = 2
        assert lru['foo1'] == 1
        assert lru['foo2'] == 2

    def test_get(self):
        lru = LRU(10)
        lru['foo1'] = 1
        assert lru.get('foo1') == 1
        assert lru.get('foo1', 2) == 1

    def test_get_default(self):
        lru = LRU(10)
        lru['foo1'] = 1
        assert lru.get('foo2') is None
        assert lru.get('foo2', 2) == 2

    def test_delitem(self):
        lru = LRU(10)
        lru['foo1'] = 1
        assert 'foo1' in lru
        del lru['foo1']
        assert 'foo1' not in lru

    def test_empty(self):
        lru = LRU(10)
        assert bool(lru) is False
        lru['foo1'] = '1'
        assert bool(lru) is True

    def test_setitem_overflow(self):
        lru = LRU(2)
        lru['foo1'] = 1
        lru['foo2'] = 2
        lru['foo3'] = 3

        assert 'foo1' not in lru
        assert 'foo2' in lru
        assert 'foo3' in lru

    def test_length(self):
        lru = LRU(2)
        assert len(lru) == 0
        lru['foo1'] = 1
        assert len(lru) == 1
        lru['foo2'] = 2
        assert len(lru) == 2
        lru['foo3'] = 3
        assert len(lru) == 2

        del lru['foo3']
        assert len(lru) == 1


class TestImmutableDictList(object):
    def test_named(self):
        res = ImmutableDictList([('one', 10), ('two', 5), ('three', 3)])
        assert res[0] == 10
        assert res[2] == 3
        assert res['one'] == 10
        assert res['three'] == 3
        assert len(res) == 3

    def test_named_iteritems(self):
        res = ImmutableDictList([('one', 10), ('two', 5), ('three', 3)])
        itr = res.iteritems()
        assert next(itr) == ('one', 10)
        assert next(itr) == ('two', 5)
        assert next(itr) == ('three', 3)
        try:
            next(itr)
        except StopIteration:
            pass
        else:
            assert False, 'StopIteration expected'



# This file is part of the MapProxy project.
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import division
import pytest
from mapproxy.util.ext.wmsparse.util import parse_datetime_range, parse_duration
from mapproxy.util.ext.wmsparse.duration import parse_datetime

expected_dim = {'time': {
    'values': '2020-09-22T11:20:00Z,2020-09-22T13:20:00Z,2020-09-22T15:20:00Z',
    'default': '2020-09-22T14:20:00Z'},
    'dim_reference_time': {
    'values': '2020-09-22T11:20:00Z,2020-09-22T13:20:00Z,2020-09-22T15:20:00Z',
    'default': '2020-09-22T14:20:00Z'}}


@pytest.fixture(scope="module")
def config_file():
    return "dimension.yaml"


class TestIsoDate(object):

    @pytest.mark.parametrize("test_input,expected_value", [

        # test_input should be string
        # expected array ==> [year,month,day,hour,mins,secs,tz]

        ("2020-08-01T12:43:38Z", [2020, 8, 1, 12, 43, 38]),

        ("1999-12-30T01:59:03Z", [1999, 12, 30, 1, 59, 3]),

    ])
    def test_parsing_datetime(self, test_input, expected_value):
        # test = "2020-08-01T12:43:38Z"
        expected_year = expected_value[0]
        expected_month = expected_value[1]
        expected_day = expected_value[2]
        expected_hour = expected_value[3]
        expected_mins = expected_value[4]
        expected_secs = expected_value[5]

        result = parse_datetime(test_input)
        test_tz = result.tzinfo
        test_tz = test_tz.tzname(test_tz)

        assert result.year == expected_year
        assert result.month == expected_month
        assert result.day == expected_day
        assert result.hour == expected_hour
        assert result.minute == expected_mins
        assert result.second == expected_secs
        assert test_tz == "UTC"

    @pytest.mark.parametrize("test_input,only_time,expected_values", [

        # test_input should be string
        # only_time flag to indicate only time parsing
        # expected array ==> [year,month,day,secs]

        ("PT1H", True, [0, 0, 0, 3600]),
        ("PT1H30M45S", True, [0, 0, 0, 5445]),
        ("P6MT", False, [0, 6, 0, 0]),
        ("P1MT", False, [0, 1, 0, 0]),
        ("P1Y2M1W3DT", False, [1, 2, 10, 0]),
    ])
    def test_parse_duration(self, test_input, only_time, expected_values):
        if only_time:
            expect = expected_values[-1]
            result = parse_duration(test_input).seconds
            assert expect == result

        else:
            expected_years = expected_values[0]
            expected_months = expected_values[1]
            expected_days = expected_values[2]
            expected_seconds = expected_values[3]

            result = parse_duration(test_input)

            result_years = int(result.years)
            result_months = int(result.months)
            result_days = int(result.days)
            result_seconds = int(result.seconds)

            assert expected_years == result_years
            assert expected_months == result_months
            assert expected_days == result_days
            assert expected_seconds == result_seconds

    @pytest.mark.parametrize("test_input,exp_values", [

        # test_input should be string
        # only_time flag to indicate only time parsing
        # exp_values array ==> [time range values]
        ("2020-09-22T00:00:00Z/2020-09-23T00:00:00Z/PT12H", ['2020-09-22T00:00:00Z',
                                                             '2020-09-22T12:00:00Z', '2020-09-23T00:00:00Z']),
        ("2020-08-01T00:00:00Z/P1DT2H30M", ['2020-08-01T00:00:00Z',
                                            '2020-08-02T02:30:00Z']),
        ("P1DT2H30M/2020-08-30T00:00:00Z", ['2020-08-28T21:30:00Z',
                                            '2020-08-30T00:00:00Z'])
    ])
    def test_parse_datetime_range(self, test_input, exp_values):

        exp_values = set(exp_values)
        result = set(parse_datetime_range(test_input))

        assert exp_values == result



# This file is part of the MapProxy project.
# Copyright (C) 2010 Omniscale <http://omniscale.de>
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import os

import pytest

from mapproxy.script.grids import grids_command
from mapproxy.test.helper import capture


FIXTURE_DIR = os.path.join(os.path.dirname(__file__), "fixture")
GRID_NAMES = ["global_geodetic_sqrt2", "grid_full_example", "another_grid_full_example"]
UNUSED_GRID_NAMES = ["GLOBAL_GEODETIC", "GLOBAL_MERCATOR", "GLOBAL_WEBMERCATOR"]


class TestUtilGrids(object):

    def setup_method(self):
        self.mapproxy_config_file = os.path.join(FIXTURE_DIR, "util_grids.yaml")
        self.args = ["command_dummy", "-f", self.mapproxy_config_file]

    def test_config_not_found(self):
        self.args = ["command_dummy", "-f", "foo.bar"]
        with capture() as (_, err):
            with pytest.raises(SystemExit):
                grids_command(self.args)
        assert err.getvalue().startswith("ERROR:")

    def test_list_configured(self):
        self.args.append("-l")
        with capture() as (out, err):
            grids_command(self.args)
        captured_output = out.getvalue()
        for grid in GRID_NAMES:
            assert grid in captured_output

        number_of_lines = sum(1 for line in captured_output.split("\n") if line)

        assert number_of_lines == len(GRID_NAMES)

    def test_list_configured_all(self):
        self.args.append("-l")
        self.args.append("--all")
        with capture() as (out, err):
            grids_command(self.args)
        captured_output = out.getvalue()
        for grid in GRID_NAMES + UNUSED_GRID_NAMES:
            assert grid in captured_output

        number_of_lines = sum(1 for line in captured_output.split("\n") if line)

        assert number_of_lines == len(UNUSED_GRID_NAMES) + len(GRID_NAMES)

    def test_display_single_grid(self):
        self.args.append("-g")
        self.args.append("GLOBAL_MERCATOR")
        with capture() as (out, err):
            grids_command(self.args)
        captured_output = out.getvalue()
        assert "GLOBAL_MERCATOR" in captured_output

    def test_ignore_case(self):
        self.args.append("-g")
        self.args.append("global_geodetic")
        with capture() as (out, err):
            grids_command(self.args)
        captured_output = out.getvalue()
        assert "GLOBAL_GEODETIC" in captured_output

    def test_all_grids(self):
        with capture() as (out, err):
            grids_command(self.args)
        captured_output = out.getvalue()
        assert "GLOBAL_MERCATOR" in captured_output
        assert "origin*: 'll'" in captured_output



# This file is part of the MapProxy project.
# Copyright (C) 2010 Omniscale <http://omniscale.de>
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
Service exception handling (WMS exceptions, XML, in_image, etc.).
"""
from mapproxy.exception import ExceptionHandler, XMLExceptionHandler
from mapproxy.response import Response
from mapproxy.image.message import message_image
from mapproxy.image.opts import ImageOptions
import mapproxy.service
from mapproxy.template import template_loader
get_template = template_loader(mapproxy.service.__package__, 'templates')


class WMSXMLExceptionHandler(XMLExceptionHandler):
    template_func = get_template


class WMS100ExceptionHandler(WMSXMLExceptionHandler):
    """
    Exception handler for OGC WMS 1.0.0 ServiceExceptionReports
    """
    template_file = 'wms100exception.xml'
    content_type = 'text/xml'


class WMS110ExceptionHandler(WMSXMLExceptionHandler):
    """
    Exception handler for OGC WMS 1.1.0 ServiceExceptionReports
    """
    template_file = 'wms110exception.xml'
    mimetype = 'application/vnd.ogc.se_xml'


class WMS111ExceptionHandler(WMSXMLExceptionHandler):
    """
    Exception handler for OGC WMS 1.1.1 ServiceExceptionReports
    """
    template_file = 'wms111exception.xml'
    mimetype = 'application/vnd.ogc.se_xml'


class WMS130ExceptionHandler(WMSXMLExceptionHandler):
    """
    Exception handler for OGC WMS 1.3.0 ServiceExceptionReports
    """
    template_file = 'wms130exception.xml'
    mimetype = 'text/xml'


class WMSImageExceptionHandler(ExceptionHandler):
    """
    Exception handler for image exceptions.
    """

    def render(self, request_error):
        request = request_error.request
        params = request.params
        format = params.format
        size = params.size
        if size is None:
            size = (256, 256)
        transparent = ('transparent' in params
                       and params['transparent'].lower() == 'true')
        bgcolor = WMSImageExceptionHandler._bgcolor(request.params)
        image_opts = ImageOptions(format=format, bgcolor=bgcolor, transparent=transparent)
        result = message_image(request_error.msg, size=size, image_opts=image_opts)
        return Response(result.as_buffer(), content_type=params.format_mime_type)

    @staticmethod
    def _bgcolor(params):
        """
        >>> WMSImageExceptionHandler._bgcolor({'bgcolor': '0Xf0ea42'})
        '#f0ea42'
        >>> WMSImageExceptionHandler._bgcolor({})
        '#ffffff'
        """
        if 'bgcolor' in params:
            color = params['bgcolor']
            if color.lower().startswith('0x'):
                color = '#' + color[2:]
        else:
            color = '#ffffff'
        return color


class WMSBlankExceptionHandler(WMSImageExceptionHandler):
    """
    Exception handler for blank image exceptions.
    """

    def render(self, request_error):
        request_error.msg = ''
        return WMSImageExceptionHandler.render(self, request_error)



# This file is part of the MapProxy project.
# Copyright (C) 2010 Omniscale <http://omniscale.de>
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import division

import pytest

try:
    from urllib.parse import quote
except ImportError:
    from urllib import quote

from mapproxy.request.wms import WMS111MapRequest
from mapproxy.test.system import SysTest
from mapproxy.test.http import mock_httpd
from mapproxy.test.image import create_tmp_image


@pytest.fixture(scope="module")
def config_file():
    return "sld.yaml"


TESTSERVER_ADDRESS = "localhost", 42423


class TestWMS(SysTest):

    @pytest.fixture(scope="class")
    def additional_files(self, base_dir):
        base_dir.join("mysld.xml").write("<sld>")

    def setup_method(self):
        self.common_map_req = WMS111MapRequest(
            url="/service?",
            param=dict(
                service="WMS",
                version="1.1.1",
                bbox="0,0,10,10",
                width="200",
                height="200",
                srs="EPSG:4326",
                format="image/png",
                styles="",
                request="GetMap",
                exceptions="xml",
            ),
        )
        self.common_wms_url = (
            "/service?styles=&srs=EPSG%3A4326&version=1.1.1&"
            "bbox=0.0,0.0,10.0,10.0&service=WMS&format=image%2Fpng&request=GetMap"
            "&width=200&height=200"
        )

    def test_sld_url(self, app):
        self.common_map_req.params["layers"] = "sld_url"
        with mock_httpd(
            TESTSERVER_ADDRESS,
            [
                (
                    {
                        "path": self.common_wms_url
                        + "&sld="
                        + quote("http://example.org/sld.xml"),
                        "method": "GET",
                    },
                    {"body": create_tmp_image((200, 200), format="png")},
                )
            ],
        ):
            resp = app.get(self.common_map_req)
            assert resp.content_type == "image/png"

    def test_sld_file(self, app):
        self.common_map_req.params["layers"] = "sld_file"
        with mock_httpd(
            TESTSERVER_ADDRESS,
            [
                (
                    {
                        "path": self.common_wms_url + "&sld_body=" + quote("<sld>"),
                        "method": "GET",
                    },
                    {"body": create_tmp_image((200, 200), format="png")},
                )
            ],
        ):
            resp = app.get(self.common_map_req)
            assert resp.content_type == "image/png"

    def test_sld_body(self, app):
        self.common_map_req.params["layers"] = "sld_body"
        with mock_httpd(
            TESTSERVER_ADDRESS,
            [
                (
                    {
                        "path": self.common_wms_url
                        + "&sld_body="
                        + quote("<sld:StyledLayerDescriptor />"),
                        "method": "POST",
                    },
                    {"body": create_tmp_image((200, 200), format="png")},
                )
            ],
        ):
            resp = app.get(self.common_map_req)
            assert resp.content_type == "image/png"



# This file is part of the MapProxy project.
# Copyright (C) 2011 Omniscale <http://omniscale.de>
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from mapproxy.util.ext.dictspec.validator import validate, ValidationError
from mapproxy.util.ext.dictspec.spec import one_off, anything, number
from mapproxy.util.ext.dictspec.spec import required

from mapproxy.config.spec import coverage, time_spec


def validate_seed_conf(conf_dict):
    """
    Validate `conf_dict` agains seed.yaml spec.
    Returns lists with errors. List is empty when no errors where found.
    """
    try:
        validate(seed_yaml_spec, conf_dict)
    except ValidationError as ex:
        return ex.errors, ex.informal_only
    else:
        return [], True


from_to_spec = {
    'from': number(),
    'to': number(),
}

seed_yaml_spec = {
    'coverages': {
        anything(): coverage,
    },
    'seeds': {
        anything(): {
            required('caches'): [str()],
            'grids': [str()],
            'coverages': [str()],
            'refresh_before': time_spec,
            'levels': one_off([int()], from_to_spec),
            'resolutions': one_off([int()], from_to_spec),
        },
    },
    'cleanups': {
        anything(): {
            required('caches'): [str()],
            'grids': [str()],
            'coverages': [str()],
            'remove_before': time_spec,
            'remove_all': bool(),
            'levels': one_off([int()], from_to_spec),
            'resolutions': one_off([int()], from_to_spec),
        }
    },
}



# This file is part of the MapProxy project.
# Copyright (C) 2010 Omniscale <http://omniscale.de>
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import division

from io import BytesIO

import pytest

from mapproxy.request.wms import WMS111MapRequest
from PIL import Image
from mapproxy.test.image import is_png, is_jpeg
from mapproxy.test.system import SysTest


@pytest.fixture(scope="module")
def config_file():
    return "seedonly.yaml"


class TestSeedOnlyWMS(SysTest):

    def setup_method(self):
        self.common_map_req = WMS111MapRequest(
            url="/service?",
            param=dict(
                service="WMS",
                version="1.1.1",
                bbox="-180,0,0,80",
                width="200",
                height="200",
                layers="wms_cache",
                srs="EPSG:4326",
                format="image/png",
                styles="",
                request="GetMap",
                transparent=True,
            ),
        )

    def test_get_map_cached(self, app, fixture_cache_data):
        resp = app.get(self.common_map_req)
        assert resp.content_type == "image/png"
        data = BytesIO(resp.body)
        assert is_png(data)
        img = Image.open(data)
        assert img.mode == "RGB"
        # cached image has more that 256 colors, getcolors -> None
        assert img.getcolors() is None

    def test_get_map_uncached(self, app):
        self.common_map_req.params["bbox"] = "10,10,20,20"
        resp = app.get(self.common_map_req)
        assert resp.content_type == "image/png"
        data = BytesIO(resp.body)
        assert is_png(data)
        img = Image.open(data)
        assert img.mode == "RGBA"
        assert img.getcolors() == [(200 * 200, (255, 255, 255, 0))]


class TestSeedOnlyTMS(SysTest):

    def test_get_tile_cached(self, app, fixture_cache_data):
        resp = app.get("/tms/1.0.0/wms_cache/0/0/1.jpeg")
        assert resp.content_type == "image/jpeg"
        data = BytesIO(resp.body)
        assert is_jpeg(data)
        img = Image.open(data)
        assert img.mode == "RGB"
        # cached image has more that 256 colors, getcolors -> None
        assert img.getcolors() is None

    def test_get_tile_uncached(self, app):
        resp = app.get("/tms/1.0.0/wms_cache/0/0/0.jpeg")
        assert resp.content_type == "image/png"
        data = BytesIO(resp.body)
        assert is_png(data)
        img = Image.open(data)
        assert img.mode == "RGBA"
        assert img.getcolors() == [(256 * 256, (255, 255, 255, 0))]
